# Quantum Inspire SDK
#
# This file contains code modified from https://github.com/ProjectQ-Framework/ProjectQ in the QIBackend class.
# The ProjectQ code is under the Apache License 2.0.
#
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import random
import sys
from collections import defaultdict
from functools import reduce
from typing import List, Dict, Iterator, Union, Optional, Tuple, Any

from projectq.cengines import BasicEngine
from projectq.meta import LogicalQubitIDTag, get_control_count
from projectq.ops import (NOT, Allocate, Barrier, Deallocate, FlushGate, H,
                          Measure, Ph, Rx, Ry, Rz, S, Sdag, Swap, T, Tdag, X,
                          Y, Z, Command, CZ, C, R, CNOT, Toffoli)
from projectq.types import Qubit
from quantuminspire.api import QuantumInspireAPI, V1_MEASUREMENT_BLOCK_INDEX
from quantuminspire.exceptions import AuthenticationError
from quantuminspire.exceptions import ProjectQBackendError
# shortcut for Controlled Phase-shift gate (CR)
CR = C(R)


class QIBackend(BasicEngine):  # type: ignore
    """ Backend for Quantum Inspire """

    def __init__(self, num_runs: int = 1024, verbose: int = 0, quantum_inspire_api: Optional[QuantumInspireAPI] = None,
                 backend_type: Optional[Union[int, str]] = None) -> None:
        """
        Initialize the Backend object.

        :param num_runs: Number of runs to collect statistics (default is 1024).
        :param verbose: Verbosity level, defaults to 0, which produces no extra output.
        :param quantum_inspire_api: Connection to QI platform, optional parameter.
        :param backend_type: Backend to use for execution.
            When no backend_type is provided, the default backend will be used.

        :raises AuthenticationError: When an authentication error occurs.
        """
        BasicEngine.__init__(self)
        self._flushed: bool = False
        """ Because engines are meant to be 'single use' by the way ProjectQ is designed,
        any additional gates received after a FlushGate triggers an exception. """
        self._clear: bool = True
        self._qasm: str = ""
        self._reset()
        self._verbose: int = verbose
        self._cqasm: str = str()
        self._measured_states: List[Dict[int, float]] = []
        self._measured_ids: List[int] = []
        self._allocation_map: List[Tuple[int, int]] = []
        self._max_qubit_id: int = -1
        self._quantum_inspire_result: Dict[str, Any] = {}
        if quantum_inspire_api is None:
            try:
                quantum_inspire_api = QuantumInspireAPI()
            except AuthenticationError as ex:
                raise AuthenticationError('Make sure you have saved your token credentials on disk or '
                                          'provide a QuantumInspireAPI instance as parameter to QIBackend') from ex
        self._quantum_inspire_api: QuantumInspireAPI = quantum_inspire_api
        self._backend_type: Dict[str, Any] = self._quantum_inspire_api.get_backend_type(backend_type)
        if num_runs < 1 or num_runs > self._backend_type["max_number_of_shots"]:
            raise ProjectQBackendError(f'Invalid number of runs (num_runs={num_runs})')
        self._num_runs: int = num_runs
        self._full_state_projection = not self._backend_type["is_hardware_backend"]
        self._is_simulation_backend = not self._backend_type["is_hardware_backend"]
        self._max_number_of_qubits: int = self._backend_type["number_of_qubits"]
        self._one_qubit_gates: Tuple[Any, ...] = self._get_one_qubit_gates()
        self._two_qubit_gates: Tuple[Any, ...] = self._get_two_qubit_gates()
        self._three_qubit_gates: Tuple[Any, ...] = self._get_three_qubit_gates()
        self._multiple_measurements = "flags" in self._backend_type and "multiple_measurement" in \
                                      self._backend_type["flags"]
        self._parallel_computing = "flags" in self._backend_type and "parallel_computing" in \
                                   self._backend_type["flags"]

    def _get_one_qubit_gates(self) -> Tuple[Any, ...]:
        allowed_operations = self._backend_type['allowed_operations']
        if len(allowed_operations) > 0:
            one_qubit_gates = []
            for gate_set in ['single_gates', 'parameterized_single_gates', 'wait', 'barrier']:
                if gate_set in allowed_operations:
                    for gate in allowed_operations[gate_set]:
                        if gate in ['x', 'y', 'z', 'h', 's', 'sdag', 't', 'tdag', 'rx', 'ry', 'rz', 'barrier']:
                            one_qubit_gates += [getattr(sys.modules[__name__], gate.capitalize())]
                        if gate == 'wait':
                            if self._verbose >= 3:
                                print(f"ProjectQ doesn't have an equivalent gate for cQASM gate '{gate}'")
        else:
            one_qubit_gates = [X, Y, Z, H, S, Sdag, T, Tdag, Rx, Ry, Rz]
        return tuple(one_qubit_gates)

    def _get_two_qubit_gates(self) -> Tuple[Any, ...]:
        allowed_operations = self._backend_type['allowed_operations']
        if len(allowed_operations) > 0:
            two_qubit_gates = []
            for gate_set in ['dual_gates', 'parameterized_dual_gates']:
                if gate_set in allowed_operations:
                    for gate in allowed_operations[gate_set]:
                        if gate in ['cz', 'cnot', 'cr']:
                            two_qubit_gates += [getattr(sys.modules[__name__], gate.upper())]
                        elif gate == 'swap':
                            two_qubit_gates += [Swap]
        else:
            two_qubit_gates = [CZ, CNOT, CR, Swap]
        return tuple(two_qubit_gates)

    def _get_three_qubit_gates(self) -> Tuple[Any, ...]:
        allowed_operations = self._backend_type['allowed_operations']
        if len(allowed_operations) > 0:
            three_qubit_gates = []
            for gate_set in ['triple_gates']:
                if gate_set in allowed_operations:
                    for gate in allowed_operations[gate_set]:
                        if gate == 'toffoli':
                            three_qubit_gates += [Toffoli]
        else:
            three_qubit_gates = [Toffoli]
        return tuple(three_qubit_gates)

    @property
    def one_qubit_gates(self) -> Tuple[Any, ...]:
        """ Return the one qubit gates as a tuple """
        return self._one_qubit_gates

    @property
    def two_qubit_gates(self) -> Tuple[Any, ...]:
        """ Return the two qubit gates as a tuple """
        return self._two_qubit_gates

    @property
    def three_qubit_gates(self) -> Tuple[Any, ...]:
        """ Return the three qubit gates as a tuple """
        return self._three_qubit_gates

    def cqasm(self) -> str:
        """ Return cQASM code that is generated last. """
        return self._cqasm

    @property
    def qasm(self) -> str:
        """ Return qasm code at any moment in the process. """
        return self._qasm

    def is_available(self, cmd: Command) -> bool:
        """
        Via this method the ProjectQ framework determines which commands (gates) are available in the backend.

        :param cmd: Command with a gate for which to check availability.

        :return: True when the gate in the command is available on the Quantum Inspire backend.
        """
        count = get_control_count(cmd)
        g = cmd.gate
        if self._verbose >= 3:
            print(f'call to is_available with cmd {cmd} (gate {g})')
        if g == NOT and count == 2:
            return Toffoli in self.three_qubit_gates
        if g == NOT and count == 1:
            return CNOT in self.two_qubit_gates
        if g == Z and count == 1:
            return CZ in self.two_qubit_gates
        if (g == R or isinstance(g, (R,))) and count == 1:
            return CR in self.two_qubit_gates
        if count != 0:
            return False
        if g in (Measure, Allocate, Deallocate):
            return True
        if g == Barrier:
            return Barrier in self.one_qubit_gates
        if g == Swap:
            return g in self.two_qubit_gates
        if g in (T, Tdag, S, Sdag, H, X, Y, Z):
            return g in self.one_qubit_gates
        if isinstance(g, (Rx, Ry, Rz)):
            one_qubit_types = []
            for gate in self.one_qubit_gates:
                if inspect.isclass(gate):
                    one_qubit_types.append(gate)
            return isinstance(g, tuple(one_qubit_types))
        if isinstance(g, Ph):
            return False

        return False

    def _reset(self) -> None:
        """ Reset qasm string.

        Reset temporary variable :attr:`qasm` to an initial value and set a flag
        to clear variables when :meth:`~._store` is called. """
        self._clear = True
        self._qasm = ""

    def _allocate_qubit(self, index_to_add: int) -> None:
        """ Allocate qubits.

        On a simulation backend it is possible to reuse qubits.
        The advantage of reusing qubits is that less
        qubits are needed in total for the algorithm.
        A source of reusing qubits is when qubits are used as ancilla bits. Ancilla bits are used to downgrade
        complicated quantum gates into simple gates by placing controls on ancilla bits or when doing quantum error
        correction. In ProjectQ, a qubit can be re-used when it is de-allocated after usage. ProjectQ sends an
        Allocate-gate for a qubit that is going to be used and a Deallocate gate for qubits that are not used anymore.

        :attr:`_allocation_map` is the store in which the administration
        is done for assigning physical qubits that are used
        with in ProjectQ to the simulation qubits, this is as they appear in the cQASM.
        :attr:`_allocation_map` stores the assignments as tuples (simulation_bit, physical_bit)
        where 'physical_bit' is requested by ProjectQ and simulation_bit is the assignment to a bit in the simulator.
        A de-allocated physical bit is registered as -1, which means the corresponding simulation bit can be re-used.

        We strive for x-to-x allocation for qubits, which means we want to allocate a physical qubit to its
        corresponding simulation qubit. We do this to respect as much as possible the qubits of the original algorithm
        in the generated cQASM for readability.

        Only when the requested physical bit is higher than the max number of bits supported by the backend, we try
        to search for an de-allocatd ancilla bit to re-use. When an ancilla is re-used, we have to reset the qubit
        which means we have to switch to non-full state projection.

        Example: When physical bit 0..4 are allocated in reversed order we would still get:

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4)

        and not:

            (0, 4), (1, 3), (2, 2), (3, 1), (4, 0)

        When bit 6 is allocated next we get:

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (6, 6)

        and when bit 6 is de-allocated again we get:

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (6, -1)

        At this point, when bit 5 is allocated we get:

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, -1)

        When a bit is allocated with an index higher than the maximum number of qubits in the simulator we try to
        allocate an earlier de-allocated bit
        When the maximum number of qubits in the simulator is 7 [0..6], allocation of bit 7 will be on position 6

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 7)

        When bit 8 is allocated next, we cannot reuse another bit, so we add it as the next in line (bit 7)

            (0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 7), (7, 8)

        """
        if self._is_simulation_backend:
            # physical bit to add cannot be allocated already
            if next(iter(x for x in self._allocation_map if x[1] == index_to_add), None) is not None:
                raise RuntimeError(f"Bit {index_to_add} is already allocated.")

            # check if the corresponding simulation bit is in the _allocation_map already,
            # we strive for x-to-x allocation, so when (x, -1) we should reuse this bit
            allocation_entry = next(iter(x for x in self._allocation_map if x[0] == index_to_add), None)
            # also take into account the maximum number of bits we may use on the backend.
            if allocation_entry is None and (index_to_add < self._max_number_of_qubits):
                # map the bit to the corresponding simulation bit
                self._allocation_map.append((index_to_add, index_to_add))
            else:
                # check if the corresponding simulation bit was de-allocated (we strive for a x-to-x allocation)
                if allocation_entry is None or allocation_entry[1] != -1:
                    # The corresponding simulation bit is not found or this is a bit in the ancilla range
                    # We look for a free spot, a previously de-allocated bit (-1)
                    allocation_entry = next(iter(x for x in self._allocation_map if x[1] == -1), None)

                if allocation_entry is None:
                    # no free spot, add a new simulation qubit
                    self._allocation_map.append((max(self._allocation_map)[0] + 1, index_to_add))
                else:
                    # we are reusing a de-allocated simulation bit, this situation turns the circuit into non-FSP
                    if self._full_state_projection:
                        self._switch_fsp_to_nonfsp()

                    # to reuse a de-allocated bit we do a prep_z first, which is better implemented as a
                    # measurement and binary controlled x-gate
                    self._qasm += f"\nmeasure q[{allocation_entry[0]}]"
                    self._qasm += f"\nc-x b[{allocation_entry[0]}], q[{allocation_entry[0]}]"
                    index = self._allocation_map.index(allocation_entry)
                    self._allocation_map[index] = (allocation_entry[0], index_to_add)

            # keep track of the maximum qubit id on simulation backend
            self._max_qubit_id = max(self._allocation_map)[0]
        else:
            # keep track of the maximum qubit id on hardware backend
            self._max_qubit_id = max(self._max_qubit_id, index_to_add)

        if self._verbose >= 1:
            print(f"_store: Allocate gate {(index_to_add,)}")
            print(f"   _allocation_map {self._allocation_map}")

    def _deallocate_qubit(self, index_to_remove: int) -> None:
        """ De-allocate qubit.

        On a simulation backend it is possible to reuse qubits.
        When a qubit is de-allocated we register ``-1`` as physical bit id in the :attr:`_allocation_map`.
        """
        if self._is_simulation_backend:
            # determine the qubits that are not de-allocated
            allocation_entry = next(iter(x for x in self._allocation_map if x[1] == index_to_remove), None)
            if allocation_entry is None:
                raise RuntimeError(f"De-allocated bit {index_to_remove} was not allocated.")

            # deallocate the corresponding simulation bit
            index = self._allocation_map.index(allocation_entry)
            self._allocation_map[index] = (allocation_entry[0], -1)

        if self._verbose >= 1:
            print(f"_store: Deallocate gate {(index_to_remove,)}")
            print(f"   _allocation_map {self._allocation_map}")

    def _logical_to_physical(self, logical_qubit_id: int) -> int:
        """Return the physical location of the qubit with the given logical id.

        :param logical_qubit_id: ID of the logical qubit whose position should be returned.

        :return:
            Physical position of logical qubit with id qb_id.
        """
        if self.main_engine.mapper is None:
            return logical_qubit_id  # no mapping

        mapping = self.main_engine.mapper.current_mapping
        if logical_qubit_id not in mapping:
            raise RuntimeError(f"Unknown qubit id {logical_qubit_id}. Please make sure "
                               f"eng.flush() was called and that the qubit "
                               f"was not eliminated during optimization.")

        return int(mapping[logical_qubit_id])

    def _physical_to_simulated(self, physical_qubit_id: int) -> int:
        """
        Return the allocated location on the simulated backend of the qubit with the given physical qubit id.

        :param physical_qubit_id: ID of the physical qubit whose position should be returned.

        :return:
            Allocated simulation bit position of physical qubit with id physical_qubit_id.
        """
        if self._is_simulation_backend:
            allocation_entry = next(iter(x for x in self._allocation_map if x[1] == physical_qubit_id), None)
            if allocation_entry is None:
                raise RuntimeError(f"Bit position in simulation backend not found for"
                                   f" physical bit {physical_qubit_id}.")
            return allocation_entry[0]

        return physical_qubit_id

    def _simulated_to_physical(self, simulated_qubit_id: int) -> int:
        """
        Return the physical qubit id given the allocated location on the simulated backend.

        :param simulated_qubit_id: the allocated location on the simulated backend.

        :return:
            Physical bit position of simulated qubit.
        """
        if self._is_simulation_backend:
            allocation_entry = next(iter(x for x in self._allocation_map if x[0] == simulated_qubit_id), None)
            if allocation_entry is None:
                raise RuntimeError(f"Physical bit position not found for"
                                   f" simulated bit {simulated_qubit_id}.")
            physical_qubit_id = allocation_entry[1]
        else:
            physical_qubit_id = simulated_qubit_id

        return physical_qubit_id

    def _physical_to_logical(self, physical_qubit_id: int) -> int:
        """
        Return the logical qubit id given the physical qubit id. The qubit id is only converted when a mapper engine is
        initialized.

        :param physical_qubit_id: the physical location of the qubit.

        :return:
            Logical bit position of physical qubit.
        """
        if self.main_engine.mapper is None:
            return physical_qubit_id  # no mapping

        mapping = self.main_engine.mapper.current_mapping
        # current_mapping is a dictionary with keys being the
        # logical qubit ids and the values being the physical ids
        for logical_id, physical_id in mapping.items():
            if int(physical_id) == physical_qubit_id:
                return int(logical_id)

        raise RuntimeError(f"Unknown physical qubit id {physical_qubit_id}. Please make sure "
                           f"eng.flush() was called and that the qubit "
                           f"was not eliminated during optimization.")

    def _add_delayed_measurements(self) -> None:
        """ Add the recorded measurements to the qasm algorithm.

        At this point, :attr:`_measured_ids` is the collection of measurement statements in the algorithm for which no
        measurement statement has been added to the qasm yet.
        For every `measured_id` a measurement statement is added.
        """
        for logical_qubit_id in self._measured_ids:
            physical_qubit_id = self._logical_to_physical(logical_qubit_id)
            sim_qubit_id = self._physical_to_simulated(physical_qubit_id)
            self._qasm += f"\nmeasure q[{sim_qubit_id}]"

    def _switch_fsp_to_nonfsp(self) -> None:
        """Switch to non-full state projection.

        We have determined that the algorithm is non-deterministic and cannot use fsp.
        For every `measured_id` a measurement statement is added.
        """
        self._add_delayed_measurements()
        self._full_state_projection = False

    def _store(self, cmd: Command) -> None:
        """Temporarily store the command cmd.

        Translates the command and stores the results in local variables.

        :param cmd: Command to store.
        """
        if self._verbose >= 3:
            print(f"_store {id(self)}: cmd {cmd}")

        if self._clear:
            self._clear = False
            self._qasm = ""
            self._measured_states = []
            self._measured_ids = []
            self._full_state_projection = not self._backend_type["is_hardware_backend"]

        gate = cmd.gate

        if gate == Deallocate:
            index_to_remove = cmd.qubits[0][0].id
            self._deallocate_qubit(index_to_remove)
            return

        if self._flushed:
            raise RuntimeError("Same instance of QIBackend used for circuit after Flush.")

        if gate == Allocate:
            index_to_add = cmd.qubits[0][0].id
            self._allocate_qubit(index_to_add)
            return

        if gate == Measure:
            assert len(cmd.qubits) == 1 and len(cmd.qubits[0]) == 1
            sim_qubit_id = self._physical_to_simulated(cmd.qubits[0][0].id)
            logical_id = None
            for t in cmd.tags:
                if isinstance(t, LogicalQubitIDTag):
                    logical_id = t.logical_qubit_id
                    break
            if self.main_engine.mapper is None:
                logical_id = cmd.qubits[0][0].id  # no mapping
            assert logical_id is not None
            self._measured_ids += [logical_id]
            # do not add the measurement statement when fsp is possible
            if not self._full_state_projection:
                self._qasm += f"\nmeasure q[{sim_qubit_id}]"
            return

        if gate == Barrier:
            qb_pos_list = [qb.id for qr in cmd.qubits for qb in qr]
            qb_str = ','.join([f"{self._physical_to_simulated(x)}" for x in qb_pos_list])
            self._qasm += f"\nbarrier q[{qb_str}]"
            return

        # When fsp is enabled and we have skipped measurements when we find a gate after these measurements,
        # switch to non-fsp and add any delayed measurement statements.
        if self._full_state_projection and len(self._measured_ids) != 0:
            self._switch_fsp_to_nonfsp()

        if gate == NOT and get_control_count(cmd) == 1:
            # this case also covers the CX controlled gate
            ctrl_pos = self._physical_to_simulated(cmd.control_qubits[0].id)
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\ncnot q[{ctrl_pos}], q[{qb_pos}]"
        elif gate == Swap:
            q0 = self._physical_to_simulated(cmd.qubits[0][0].id)
            q1 = self._physical_to_simulated(cmd.qubits[1][0].id)
            self._qasm += f"\nswap q[{q0}], q[{q1}]"
        elif gate == X and get_control_count(cmd) == 2:
            ctrl_pos1 = self._physical_to_simulated(cmd.control_qubits[0].id)
            ctrl_pos2 = self._physical_to_simulated(cmd.control_qubits[1].id)
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\ntoffoli q[{ctrl_pos1}], q[{ctrl_pos2}], q[{qb_pos}]"
        elif gate == Z and get_control_count(cmd) == 1:
            ctrl_pos = self._physical_to_simulated(cmd.control_qubits[0].id)
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\ncz q[{ctrl_pos}], q[{qb_pos}]"
        elif isinstance(gate, (Rz, R)) and get_control_count(cmd) == 1:
            ctrl_pos = self._physical_to_simulated(cmd.control_qubits[0].id)
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            gate_name = 'cr'
            self._qasm += f"\n{gate_name} q[{ctrl_pos}],q[{qb_pos}],{gate.angle:.12f}"
        elif isinstance(gate, (Rx, Ry)) and get_control_count(cmd) == 1:
            raise NotImplementedError("controlled Rx or Ry gate not implemented")
        elif isinstance(gate, (Rx, Ry, Rz)):
            assert get_control_count(cmd) == 0
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            gate_name = str(gate)[0:2].lower()
            self._qasm += f"\n{gate_name} q[{qb_pos}],{gate.angle:.12g}"
        elif gate == Tdag and get_control_count(cmd) == 0:
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\ntdag q[{qb_pos}]"
        elif gate == Sdag and get_control_count(cmd) == 0:
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\nsdag q[{qb_pos}]"
        elif isinstance(gate, tuple(type(gate_in_set) for gate_in_set in (X, Y, Z, H, S, T))):
            assert get_control_count(cmd) == 0
            gate_str = str(gate).lower()
            qb_pos = self._physical_to_simulated(cmd.qubits[0][0].id)
            self._qasm += f"\n{gate_str} q[{qb_pos}]"
        else:
            raise NotImplementedError(f"cmd '{(cmd,)}' not implemented")

    def get_probabilities(self, qureg: List[Qubit]) -> Dict[str, float]:
        """Return the list of basis states with corresponding probabilities.

        The measured bits are ordered according to the supplied quantum
        register, i.e., the left-most bit in the state-string corresponds to
        the first qubit in the supplied quantum register.

        .. warning::
            Only call this function after the circuit has been executed!

        :param qureg: Quantum register of size n determining the contents of the probability states.

        :return:
            Dictionary mapping n-bit strings of ``0`` and ``1`` to probabilities.

        :raises RuntimeError: If no data is available (i.e., if the circuit has
                not been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).
        """
        if len(self._measured_states) == 0:
            raise RuntimeError("Please, run the circuit first!")
        mask_bits = map(lambda qubit: self._physical_to_simulated(self._logical_to_physical(qubit.id)), qureg)

        filtered_states = QIBackend._filter_histogram(self._measured_states[V1_MEASUREMENT_BLOCK_INDEX], mask_bits)

        probability_dict = {self._map_state_to_bit_string(state, qureg): probability
                            for state, probability in filtered_states.items()}

        return probability_dict

    def get_probabilities_multiple_measurement(self, qureg: List[Qubit]) -> List[Dict[str, float]]:
        """For each measurement block return the basis states with corresponding probabilities.

        The measured bits are ordered according to the supplied quantum
        register, i.e., the left-most bit in the state-string corresponds to
        the first qubit in the supplied quantum register.

        .. warning::
            Only call this function after the circuit has been executed!

        :param qureg: Quantum register of size n determining the contents of the probability states.

        :return:
            List of dictionaries mapping n-bit strings of ``0`` and ``1`` to probabilities for each measurement block.

        :raises RuntimeError: If no data is available (i.e., if the circuit has
                not been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).
        """
        if len(self._measured_states) == 0:
            raise RuntimeError("Please, run the circuit first!")

        dict_list: List[Dict[str, float]] = []

        for measured_states in self._measured_states:
            mask_bits = map(lambda qubit: self._physical_to_simulated(self._logical_to_physical(qubit.id)), qureg)
            filtered_states = QIBackend._filter_histogram(measured_states, mask_bits)

            probability_dict = {self._map_state_to_bit_string(state, qureg): probability
                                for state, probability in filtered_states.items()}
            dict_list.append(probability_dict)

        return dict_list

    def _map_state_to_bit_string(self, state: int, qureg: List[Qubit]) -> str:
        """ Map the state to a bit string

        :param state: state represented as an integer number.
        :param qureg: list of qubits for which to extract the state bit.

        :return:
            A string of ``0`` and ``1`` corresponding to the bit value in state of each Qubit in qureg.

        Example:

        .. code-block::

            >>> state = int('0b101010', 2)
            >>> qureg = [Qubit(0), Qubit(1), Qubit(5)]
            >>> print(self._map_state_to_bit_string(state, qureg)
            011

        """
        mapped_state = ''

        for qubit in qureg:
            logical_id = qubit.id
            physical_qubit_id = self._logical_to_physical(logical_id)
            sim_qubit_id = self._physical_to_simulated(physical_qubit_id)
            if int(state) & (1 << sim_qubit_id):
                mapped_state += '1'
            else:
                mapped_state += '0'

        return mapped_state

    def _run(self) -> None:
        """ Run the circuit.

        Send the circuit via the Quantum Inspire API.
        """
        if self._qasm == "":
            return

        self._finalize_qasm()
        self._execute_cqasm()
        self._filter_result_by_measured_qubits()
        self._register_random_measurement_outcome()
        self._reset()

    def _finalize_qasm(self) -> None:
        """ Finalize qasm (add version and qubits line). """
        qasm = f"version 1.0\n# cQASM generated by Quantum Inspire {self.__class__} class\n" \
               f"qubits {self._number_of_qubits}\n"
        qasm += self._qasm

        if self._verbose >= 2:
            print(qasm)
        self._cqasm = qasm

    def _execute_cqasm(self) -> None:
        """ Execute self._cqasm through the API.

        Sets :attr:`_quantum_inspire_result` with the result object in the API response.

        :raises ProjectQBackendError: When `raw_text` in result from API is not empty (indicating a backend error).
        """
        self._quantum_inspire_result = self._quantum_inspire_api.execute_qasm(
            self._cqasm,
            number_of_shots=self._num_runs,
            backend_type=self._backend_type,
            full_state_projection=self._full_state_projection
        )

        if not self._quantum_inspire_result.get("histogram", [{}])[0]:
            raw_text = self._quantum_inspire_result.get("raw_text", "no raw_text in result structure")
            raise ProjectQBackendError(
                f"Result from backend contains no histogram data!\n{raw_text}")

    def _filter_result_by_measured_qubits(self) -> None:
        """ Filters the raw result by collapsing states so that unmeasured qubits are ignored.

        Populates :attr:`_measured_states` by filtering :attr:`_quantum_inspire_result['histogram']` of the
        measurement blocks based on :attr:`_measured_ids` (which are supposed to be logical qubit id's).
        """
        nr_of_measurement_blocks = len(self._quantum_inspire_result["measurement_mask"])
        for measurement_block_index in range(nr_of_measurement_blocks):
            result_histogram = self._quantum_inspire_result["histogram"][measurement_block_index]
            mask_bits = self._get_measured_qubit_iterator(measurement_block_index)

            histogram: Dict[int, float] = {int(k): v for k, v in result_histogram.items()}
            self._measured_states.append(QIBackend._filter_histogram(histogram, mask_bits))

    @staticmethod
    def _filter_histogram(histogram: Dict[int, float], mask_bits: Iterator[int]) -> Dict[int, float]:
        """ Filter a histogram by mask_bits.

        :param histogram: input histogram mapping state to probability.
        :param mask_bits: list of bits that are to be kept in the filtered histogram.

        :return:
            Collapsed histogram mapping state to probability.

        Keys in a histogram dict are the states represented as an integer number (may be int or string), values the
        probability corresponding to that state.

        The `mask_bits` list specifies the relevant bits, any bit not set to 1 in the mask will be ignored (masked out).
        The probabilities of equivalent states summed.

        For example, if we have two states ``0b0010`` and ``0b0011``, and mask_bits is ``[1]``
        we only care about bit 1 (the second bit from the right).
        The output will specify only one state, ``0b0010``, and the probability is the sum of the
        two original probabilities.

        """
        mask = reduce(lambda x, y: x | (1 << y), mask_bits, 0)

        filtered_states: Dict[int, float] = defaultdict(lambda: 0)
        for state, probability in histogram.items():
            filtered_states[state & mask] += probability

        return dict(filtered_states)

    def _register_random_measurement_outcome(self) -> None:
        """
        Samples the :attr:`_measured_states` for a single result for the last measurement block
        and registers this as the outcome of the circuit. """

        class QB:
            def __init__(self, qubit_id: int) -> None:
                self.id: int = qubit_id

        random_measurement = self._sample_measured_states_once()
        simulated_bits = self._get_measured_qubit_iterator(V1_MEASUREMENT_BLOCK_INDEX)

        for sim_qubit_id in simulated_bits:
            result = bool(random_measurement & (1 << sim_qubit_id))

            logical_qubit_id = self._physical_to_logical(self._simulated_to_physical(sim_qubit_id))
            self.main_engine.set_measurement_result(QB(logical_qubit_id), result)

    def _get_measured_qubit_iterator(self, measurement_block_index: int) -> Iterator[int]:
        """
        Get an iterator for the measured qubits. The iterator returns the indexes of the qubits that are measured.
        With fsp, the simulator will return all qubits as measured, but when there were measurements in the algorithm
        we want to get these measured qubits only.

        :param measurement_block_index: measurement block index for multi measurement results.

        :return:
            Iterator which iterated the indexes of the measured qubits
        """
        if self._full_state_projection and len(self._measured_ids) != 0:
            bit_iterator = map(lambda bit: self._physical_to_simulated(self._logical_to_physical(bit)),
                               self._measured_ids)
        else:
            measured_mask = self._quantum_inspire_result["measurement_mask"][measurement_block_index]
            bit_iterator = map(lambda bit_id: bit_id,
                               [bit_id for bit_id in range(len(measured_mask)) if measured_mask[bit_id] == 1])
        return bit_iterator

    def _sample_measured_states_once(self) -> int:
        """
        Obtain a random state for the last measurement block from the :attr:`_measured_states`, taking into account
        the probability distribution.
        """
        states = list(self._measured_states[V1_MEASUREMENT_BLOCK_INDEX].keys())
        weights = list(self._measured_states[V1_MEASUREMENT_BLOCK_INDEX].values())
        return random.choices(states, weights=weights)[0]

    @property
    def _number_of_qubits(self) -> int:
        """ Return the number of qubits used in the circuit. If it is a hardware backend return max nr of qubits """
        if self._is_simulation_backend:
            return self._max_qubit_id + 1

        return self._max_number_of_qubits

    def receive(self, command_list: List[Command]) -> None:
        """
        Receives a command list and, for each command, stores it until completion.

        :param command_list: List of commands to execute.
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._store(cmd)
            else:
                self._run()
                self._flushed = True
                self._reset()



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module credentials
==================

The following functions use a resource file to store credentials information
for the user. The default location of this resource file is
:file:`.quantuminspire/qirc` in the user's home directory.
This default location is indicated with `DEFAULT_QIRC_FILE` in the following function signatures.

.. autofunction:: load_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]
.. autofunction:: read_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]
.. autofunction:: store_account(token: str, filename: str = DEFAULT_QIRC_FILE, overwrite: bool = False) -> None
.. autofunction:: delete_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None
.. autofunction:: save_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None
.. autofunction:: enable_account(token: str) -> None:
.. autofunction:: get_token_authentication(token: Optional[str] = None) -> TokenAuthentication:
.. autofunction:: get_basic_authentication(email: str, password: str) -> BasicAuthentication:
.. autofunction:: get_authentication() -> Union[TokenAuthentication, BasicAuthentication]:

"""

from getpass import getpass
import json
import os
from typing import Optional, Union
import warnings

from coreapi.auth import BasicAuthentication, TokenAuthentication

DEFAULT_QIRC_FILE = os.path.join(os.path.expanduser("~"), '.quantuminspire', 'qirc')


def load_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]:
    """ Try to load an earlier stored Quantum Inspire token from file or environment.

    Load the token when found. This method looks for the token in two locations, in the following order:

        1. In the environment variable (:envvar:`QI_TOKEN`).
        2. In the file with `filename` given or, when not given, the default resource file
           :file:`.quantuminspire/qirc` in the user's home directory.

    :param filename: full path to the resource file. If no `filename` is given, the default resource file
        :file:`.quantuminspire/qirc` in the user's home directory is used.
    :return:
        The Quantum Inspire token or None when no token is found.
    """
    token = os.environ.get('QI_TOKEN', None) or read_account(filename)
    return token


def read_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]:
    """ Try to read an earlier stored Quantum Inspire token from file.

    his method looks for the token in the file with `filename` given or,
    when no `filename` is given, the default resource file :file:`.quantuminspire/qirc` in the user's home directory.

    :param filename: full path to the resource file. If no filename is given, the default resource file
        :file:`.quantuminspire/qirc` in the user's home directory is used.

    :return:
        The Quantum Inspire token or None when no token is found or token is empty.
    """
    try:
        with open(filename, 'r', encoding='utf-8') as file:
            accounts = json.load(file)
            token: Optional[str] = accounts['token']
    except (OSError, KeyError, ValueError):  # file does not exist or is empty/invalid
        token = None
    return token if token else None


def store_account(token: str, filename: str = DEFAULT_QIRC_FILE, overwrite: bool = False) -> None:
    """Store the token in a resource file.

    Replace an existing token only when overwrite=True.

    :param token: the Quantum Inspire token to store to disk.
    :param filename: full path to the resource file. If no `filename` is given, the default resource file
        :file:`.quantuminspire/qirc` in the user's home directory is used.
    :param overwrite: overwrite an existing token.
    """
    stored_token = read_account(filename)
    if stored_token and stored_token != token and not overwrite:
        warnings.warn('Token already present. Set overwrite=True to overwrite.')
        return
    save_account(token, filename)


def delete_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None:
    """Remove the token from the resource file.

    :param token: the Quantum Inspire token to remove.
    :param filename: full path to the resource file. If no `filename` is given, the default resource file
        :file:`.quantuminspire/qirc` in the user's home directory is used.
    """
    stored_token = read_account(filename)
    if stored_token == token:
        save_account('', filename)


def save_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None:
    """Save the token to a file.

    Save the token to the file with `filename` given, otherwise save to the default resource file.
    An existing token is overwritten. Use :meth:`~.store_account` to prevent overwriting an existing token.

    :param token: the Quantum Inspire token to save.
    :param filename: full path to the resource file. If no `filename` is given, the default resource file
        :file:`.quantuminspire/qirc` in the user's home directory is used.
    """
    accounts = {'token': token}
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w', encoding='utf-8') as config_file:
        json.dump(accounts, config_file, indent=2)


def enable_account(token: str) -> None:
    """Save the token to the internal environment, that will be used by load_account for the session.

    When a token was already loaded from the system environment it is overwritten.
    The system environment is not effected.

    :param token: the Quantum Inspire token to be used by :meth:`~.load_account` for the session.
    """
    os.environ['QI_TOKEN'] = token


def get_token_authentication(token: Optional[str] = None) -> TokenAuthentication:
    """Set up token authentication for Quantum Inspire to be used in the API.

    :param token: the Quantum Inspire token to set in TokenAuthentication. When no token is given,
        the token returned from :meth:`~.load_account` is used.

    :return:
        The token authentication for Quantum Inspire.
    """
    if not token:
        token = load_account()
    return TokenAuthentication(token, scheme="token")


def get_basic_authentication(email: str, password: str) -> BasicAuthentication:
    """Set up basic authentication for Quantum Inspire to be used in the API.

    :param email: a valid email address.
    :param password: password for the account.

    :return:
        The basic authentication for Quantum Inspire.
    """
    return BasicAuthentication(email, password)


def get_authentication() -> Union[TokenAuthentication, BasicAuthentication]:
    """ Gets the authentication for connecting to the Quantum Inspire API.

        First it tries to load a token, saved earlier. When a token is not found it tries to login
        with basic authentication read from the environment variables QI_EMAIL and QI_PASSWORD. When the environment
        variables are not both set, email and password are read from standard input.

    :return:
        The token or basic authentication for Quantum Inspire.
    """
    token = load_account()
    if token is not None:
        return get_token_authentication(token)

    email = os.environ.get('QI_EMAIL', None)
    password = os.environ.get('QI_PASSWORD', None)
    if email is None or password is None:
        print('Enter email:')
        email = input()
        print('Enter password')
        password = getpass()

    return get_basic_authentication(email, password)



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Union
from qiskit.exceptions import QiskitError
from qiskit.result import postprocess, Result
from qiskit.result.models import ExperimentResult

from quantuminspire.exceptions import QiskitBackendError


class QIResult(Result):  # type: ignore
    """
    A result object returned by QIJob:
        qi_backend = QI.get_backend('QX single-node simulator')
        job = qi_backend.retrieve_job(job_id)
        qi_result = job.result()
    """
    def __init__(self, backend_name: str, backend_version: str, qobj_id: str, job_id: str, success: bool,
                 results: List[ExperimentResult], date: Any = None, status: Any = None, header: Any = None,
                 **kwargs: Any) -> None:
        """
        Construct a new QIResult object. Not normally called directly, use a QIJob to get the QIResult.
        Based on Qiskit Result.

        :param backend_name: backend name.
        :param backend_version: backend version, in the form X.Y.Z.
        :param qobj_id: user-generated Qobj id.
        :param job_id: unique execution id from the backend.
        :param success: True if complete input qobj executed correctly. (Implies each experiment success)
        :param results: corresponding results for array of experiments of the input qobj
        :param date: date to be added to the result object
        :param status: status to be added to the result object
        :param header: header to be added to the result object
        :param kwargs: other parameters (added as metadata to the result object)
        """
        super().__init__(backend_name, backend_version, qobj_id, job_id, success,
                         results, date, status, header, **kwargs)

    def get_raw_result(self, field_name: str, experiment: Any = None) -> Union[List[Dict[str, Any]],
                                                                               List[List[List[str]]],
                                                                               List[List[Dict[str, float]]]]:
        """
        Get the specific and unprocessed result data of an experiment.
        Can handle single and multi measurement results.

        :param field_name: the specific result that is requested
            Can be one of 'calibration', 'counts', 'memory', 'probabilities' or the multi measurement results
            'counts_multiple_measurement', 'memory_multiple_measurement', 'probabilities_multiple_measurement'

        :param experiment: the index of the experiment (str or QuantumCircuit or Schedule or int or None),
            as specified by ``get_data()``.

        :return:
            A list with result structures which holds the specific unprocessed result for each experiment.

        :raises QiskitBackendError: raised if the results requested are not in the results for the experiment(s).
        """
        results_list = []

        if experiment is None:
            exp_keys = range(len(self.results))
        else:
            exp_keys = [experiment]  # type: ignore

        for key in exp_keys:
            if field_name in self.data(key).keys():
                result_values = self.data(key)[field_name]
                results_list.append(result_values)
            else:
                raise QiskitBackendError(f'Result does not contain {field_name} data for experiment "{key}"')
        return results_list

    def get_probabilities(self, experiment: Any = None) -> Union[Dict[str, float], List[Dict[str, float]]]:

        """Get the probability data of an experiment. The probability data is added as a separate result by
        Quantum Inspire backend.

        :param experiment: the index of the experiment (str or QuantumCircuit or Schedule or int or None),
            as specified by ``get_data()``.

        :return:
            A single or a list of dictionaries which holds the states and probabilities for respectively 1 or more
            experiment result.

        :raises QiskitBackendError: raised if there are no probabilities in a result for the experiment(s).
        """
        if experiment is None:
            exp_keys = range(len(self.results))
        else:
            exp_keys = [experiment]  # type: ignore

        dict_list: List[Dict[str, float]] = []
        for key in exp_keys:
            exp = self._get_experiment(key)
            try:
                header = exp.header.to_dict()
            except (AttributeError, QiskitError):  # header is not available
                header = None

            if "probabilities" in self.data(key).keys():
                probabilities = self.data(key)["probabilities"]
                dict_list.append(postprocess.format_counts(probabilities, header))
            else:
                raise QiskitBackendError(f'No probabilities for experiment "{key}"')

        # Return first item of dict_list if size is 1
        if len(dict_list) == 1:
            return dict_list[0]

        return dict_list

    def get_probabilities_multiple_measurement(self, experiment: Any = None) -> Union[List[Dict[str, float]],
                                                                                      List[List[Dict[str, float]]]]:
        """
        Get the probability data of an experiment for all measurement blocks.
        The probability data is added as a separate result by Quantum Inspire backend.

        :param experiment: the index of the experiment (str or QuantumCircuit or Schedule or int or None),
            as specified by ``get_data()``.

        :return:
            One list or a list of list of dictionaries which holds the states and probabilities for each measurement
            block for respectively 1 or more experiment result.

        :raises QiskitBackendError: raised if there are no multi measurement probabilities in a result for the
            experiment(s).
        """
        if experiment is None:
            exp_keys = range(len(self.results))
        else:
            exp_keys = [experiment]  # type: ignore

        list_of_dict_list: List[List[Dict[str, float]]] = []

        for key in exp_keys:
            exp = self._get_experiment(key)
            try:
                header = exp.header.to_dict()
            except (AttributeError, QiskitError):  # header is not available
                header = None

            if "probabilities_multiple_measurement" in self.data(key).keys():
                dict_list: List[Dict[str, float]] = []
                for probabilities in self.data(key)["probabilities_multiple_measurement"]:
                    dict_list.append(postprocess.format_counts(probabilities, header))
                list_of_dict_list.append(dict_list)
            else:
                raise QiskitBackendError(f'No probabilities_multiple_measurement for experiment "{key}"')

        # Return first item of list_dict_list if size is 1
        if len(list_of_dict_list) == 1:
            return list_of_dict_list[0]

        return list_of_dict_list

    def get_calibration(self, experiment: Any = None) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Get the calibration data of an experiment. The calibration data is added as a separate result item by
        Quantum Inspire backend.

        :param experiment: the index of the experiment, as specified by ``get_data()``.
                experiment can be: str or QuantumCircuit or Schedule or int or None

        :return:
            Single or a list of dictionaries which holds the calibration data for respectively 1 or more experiment(s).
            Exact format depends on the backend. A simulator backend has no calibration data (None is returned)

        :raises QiskitBackendError: raised if there is no calibration data in a result for the experiment(s).
        """
        if experiment is None:
            exp_keys = range(len(self.results))
        else:
            exp_keys = [experiment]  # type: ignore

        dict_list: List[Dict[str, float]] = []
        for key in exp_keys:
            if "calibration" in self.data(key).keys():
                calibration = self.data(key)["calibration"]
                dict_list.append(calibration)
            else:
                raise QiskitBackendError(f'No calibration data for experiment "{key}"')

        # Return first item of dict_list if size is 1
        if len(dict_list) == 1:
            return dict_list[0]

        return dict_list



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from typing import List, Optional, Any, Dict

import coreapi
from qiskit.providers.models import QasmBackendConfiguration
from qiskit.providers.provider import ProviderV1 as Provider

from quantuminspire.api import QuantumInspireAPI, QI_URL
from quantuminspire.credentials import get_token_authentication, get_basic_authentication
from quantuminspire.exceptions import QiskitBackendError
from quantuminspire.qiskit.backend_qx import QuantumInspireBackend


class QuantumInspireProvider(Provider):  # type: ignore
    """ Provides a backend and an api for a single Quantum Inspire account. """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._backends: Optional[List[QuantumInspireBackend]] = None
        self._filter: Optional[str] = None
        self._api: Optional[QuantumInspireAPI] = None

    def __str__(self) -> str:
        return 'QI'

    def get_api(self) -> QuantumInspireAPI:
        """ Provides the authenticated Quantum Inspire API.

        :return:
            The authenticated API.
        """
        if self._api is None:
            raise QiskitBackendError('Authentication details have not been set.')

        return self._api

    def backends(self, name: Optional[str] = None, **kwargs: Any) -> List[QuantumInspireBackend]:
        """ Provides a list of backends.

        :param name: Name of the requested backend. When None we take all available backends
        :param kwargs: Used for filtering, not implemented.

        :return:
            List of backends that meet the filter requirements (is_allow == True).
        """
        if self._backends is not None and self._filter == name:
            return self._backends

        if self._api is None:
            raise QiskitBackendError('Authentication details have not been set.')

        available_backends = self._api.get_backend_types()
        if name is not None:
            available_backends = list(filter(lambda b: b['name'] == name, available_backends))
        self._filter = name
        self._backends = []
        for backend in available_backends:
            if backend['is_allowed']:
                config = copy(QuantumInspireBackend.DEFAULT_CONFIGURATION)
                self._adjust_backend_configuration(config, backend)
                self._backends.append(QuantumInspireBackend(self._api, provider=self, configuration=config))

        return self._backends

    @staticmethod
    def _adjust_backend_configuration(config: QasmBackendConfiguration, backend: Dict[str, Any]) -> None:
        """
        Overwrites the default configuration with the backend dependent configuration items.

        :param config: The configuration with default configuration that is adjusted.
        :param backend: The backend for which the configuration items are adjusted.
        """
        config.backend_name = backend['name']
        config.n_qubits = backend['number_of_qubits']
        if len(backend['allowed_operations']) > 0:
            config.basis_gates = []
            for keys in backend['allowed_operations']:
                if keys in ['single_gates', 'parameterized_single_gates', 'dual_gates',
                            'parameterized_dual_gates', 'triple_gates', 'wait', 'barrier']:
                    for gate in backend['allowed_operations'][keys]:
                        if gate in ['x', 'y', 'z', 'h', 's', 't', 'rx', 'ry', 'rz', 'swap', 'cz', 'barrier']:
                            config.basis_gates += [gate]
                        elif gate == 'wait':
                            config.basis_gates += ['delay']
                        elif gate == 'i':
                            config.basis_gates += ['id']
                        elif gate == 'sdag':
                            config.basis_gates += ['sdg']
                        elif gate == 'tdag':
                            config.basis_gates += ['tdg']
                        elif gate == 'cnot':
                            config.basis_gates += ['cx']
                        elif gate == 'toffoli':
                            config.basis_gates += ['ccx']
            if 'rz' in config.basis_gates and 'ry' in config.basis_gates:
                config.basis_gates += ['p']

        config.simulator = not backend['is_hardware_backend']
        config.conditional = not backend['is_hardware_backend']
        config.max_shots = backend['max_number_of_shots']
        max_experiments = backend['max_number_of_simultaneous_jobs']
        config.max_experiments = max_experiments if max_experiments else 1
        coupling_map = []
        for i in range(len(backend['topology']['edges'])):
            for j in backend['topology']['edges'][i]:
                coupling_map.append([i, j])
        config.coupling_map = None if len(coupling_map) == 0 else coupling_map
        config.multiple_measurements = "flags" in backend and "multiple_measurement" in backend["flags"]
        config.parallel_computing = "flags" in backend and "parallel_computing" in backend["flags"]

    def set_authentication_details(self, email: str, password: str, qi_url: str = QI_URL) -> None:
        """Set a single authentication for Quantum Inspire.

        .. deprecated:: 0.5.0
           Replaced with method :meth:`~.set_basic_authentication`

        :param email: A valid email address.
        :param password: Password for the account.
        :param qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.

        """
        self.set_basic_authentication(email, password, qi_url)

    def set_basic_authentication(self, email: str, password: str, qi_url: str = QI_URL) -> None:
        """Set up basic authentication for Quantum Inspire.

        :param email: A valid email address.
        :param password: Password for the account.
        :param qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        """
        authentication = get_basic_authentication(email, password)
        self.set_authentication(authentication, qi_url)

    def set_token_authentication(self, token: str, qi_url: str = QI_URL) -> None:
        """
        Set up token authentication for Quantum Inspire.

        :param token: A valid token.
        :param qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        """
        authentication = get_token_authentication(token)
        self.set_authentication(authentication, qi_url)

    def set_authentication(self, authentication: Optional[coreapi.auth.AuthBase] = None,
                           qi_url: str = QI_URL, project_name: Optional[str] = None) -> None:
        """
        Initializes the API and sets the authentication for Quantum Inspire.

        :param authentication: The authentication, can be one of the following coreapi authentications:

            * ``BasicAuthentication(email, password)``, HTTP authentication with valid email/password.
            * ``TokenAuthentication(token, scheme="token")``, token authentication with a valid API-token.

            When authentication is ``None``, the api will try to load a token from the default resource.

        :param qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        :param project_name: The project name of the project that is used for executing the jobs.
        """
        self._api = QuantumInspireAPI(qi_url, authentication, project_name)

    def set_project_name(self, project_name: str) -> None:
        """
        Set the project name for the Quantum Inspire project that is used for executing the jobs.
        Default a generated project-name is used, but it can be overridden by the user.

        :param project_name: project name of the QI project used for executing the jobs.
        """
        if self._api is not None:
            self._api.project_name = project_name



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, Dict, List

from qiskit.qobj import QasmQobjExperiment
from quantuminspire.exceptions import QiskitBackendError


class Measurements:
    def __init__(self) -> None:
        """
        The Measurements class registers measurements and the translations of measurements of qubits to classical
        registers.
        _measurements_reg and _measurements_state are list of lists, for each measurement the lists contains
        a list of [qubit_index, classical_bit_index], which represents the measurement of a qubit to a
        classical bit.
        _measurements_reg contains the bit indices of the qubit and classical bit as used in the algorithm
        _measurements_state contains the positional indexes in the resulting qubit_state and classical_state arrays
        """
        self._number_of_qubits = 0
        self._number_of_clbits = 0
        self._measurements_reg: List[List[int]] = []
        self._measurements_state: List[List[int]] = []

    @property
    def nr_of_qubits(self) -> int:
        """
        :return: Return number of qubits in the algorithm
        """
        return self._number_of_qubits

    @nr_of_qubits.setter
    def nr_of_qubits(self, nr_of_qubits: int) -> None:
        """
        Setter for _number_of_qubits
        """
        self._number_of_qubits = nr_of_qubits

    @property
    def nr_of_clbits(self) -> int:
        """
        :return: Return number of classical bits in the algorithm
        """
        return self._number_of_clbits

    @nr_of_clbits.setter
    def nr_of_clbits(self, nr_of_clbits: int) -> None:
        """
        Setter for _number_of_clbits
        """
        self._number_of_clbits = nr_of_clbits

    @property
    def measurements_reg(self) -> List[List[int]]:
        """
        :return: Return the bit indices of the qubit and classical bit as used in the algorithm
        """
        return self._measurements_reg

    @measurements_reg.setter
    def measurements_reg(self, measurements_reg: List[List[int]]) -> None:
        """
        Setter for _measurements_reg
        """
        self._measurements_reg = measurements_reg

    @property
    def measurements_state(self) -> List[List[int]]:
        """
        :return: Return the positional indexes in the resulting qubit_state and classical_state arrays
        """
        return self._measurements_state

    @measurements_state.setter
    def measurements_state(self, measurements_state: List[List[int]]) -> None:
        """
        Setter for _measurements_state
        """
        self._measurements_state = measurements_state

    @classmethod
    def from_experiment(cls, experiment: QasmQobjExperiment) -> Measurements:
        """ Determines the measured qubits and classical bits for an experiment.

        The full-state measured qubits is returned when no measurements are present in the compiled circuit.

        .. code::

            q = QuantumRegister(3)
            c = ClassicalRegister(3)
            circuit = QuantumCircuit(q, c)

            circuit.measure(2, 0)

        for this measurement [2, 0] is added to the list _measurements_reg
        for this measurement [0, 2] is added to the list _measurements_state, qubit 2 is located at position 0
        in the resulting state array and can be indexed with [0]. classical 0 is located at position 2
        in the resulting state array and can be indexed with [2].

        :param experiment: The experiment with gate operations and header.

        :return: Instance of Measurements with collected measurements from experiment
        """
        instance = cls()
        header = experiment.header
        instance.nr_of_qubits = header.n_qubits
        instance.nr_of_clbits = header.memory_slots

        for instruction in experiment.instructions:
            if instruction.name == 'measure':
                instance.measurements_reg.append([instruction.qubits[0], instruction.memory[0]])
                instance.measurements_state.append([instance.nr_of_qubits - 1 - instruction.qubits[0],
                                                    instance.nr_of_clbits - 1 - instruction.memory[0]])

        if not instance.measurements_reg:
            instance.measurements_reg = [[index, index] for index in range(instance.nr_of_qubits)]
            instance.measurements_state = instance.measurements_reg

        # do some validations
        instance.validate_number_of_clbits()

        return instance

    @property
    def max_measurement_index(self) -> int:
        """
        :return: Return the highest classical bit that is used as a storage for a qubit measurement
        """
        return max(measurement[1] for measurement in self._measurements_reg)

    def get_qreg_for_conditional_creg(self, creg: int) -> int:
        """
        This method returns the qubit register that was measured to be stored in the classical register given

            1. Try to find a measurement where the classical bit creg is used as storage -> return the qubit index of
            this measurement.
            2. When no measurement is found for this classical bit, we assume the equivalent qubit (same index) is used
            First check if a measurement is found for this qubit to another classical register, we raise an error.
            This situation is not supported.

        :param creg: The classical register.

        :return: Returns the qubit index for the qubit that is measured to the classical register creg
        """
        for q1, c1 in self._measurements_reg:
            if c1 == creg:
                return q1

        for q1, c1 in self._measurements_reg:
            if q1 == creg:
                raise QiskitBackendError(f"Classical bit {creg} used in a conditional gate is not measured "
                                         f"and the equivalent qubit {q1} is measured to another classical bit {c1}")

        return creg

    def to_dict(self) -> Dict[str, Any]:
        """
        Translate the class instance to a (minimal) dictionary used for interpreting the results from the backend

        :return:
            Dict containing the measurement list with positional indexes in the resulting qubit_state and
            classical_state and the number of classical bits.
        """
        return {'measurements_state': self._measurements_state,
                'measurements_reg': self._measurements_reg,
                'number_of_qubits': self._number_of_qubits,
                'number_of_clbits': self._number_of_clbits}

    @classmethod
    def from_dict(cls, measurement_input: Dict[str, Any]) -> Measurements:
        """
        Translate the input dictionary to an instance of class Measurements

        :param measurement_input: A dictionary with the measurement information. See method to_dict

        :return: Instance of Measurements with input from dictionary
        """
        instance = cls()
        instance.measurements_state = measurement_input['measurements_state']
        instance.measurements_reg = measurement_input['measurements_reg']
        instance.nr_of_qubits = measurement_input['number_of_qubits']
        instance.nr_of_clbits = measurement_input['number_of_clbits']

        return instance

    def validate_number_of_clbits(self) -> None:
        """ Validate the (number of) classical bits used in the measurements are valid for the algorithm

        Checks whether the number of classical bits has a value cQASM can support.

        1.  When number of classical bits is less than 1 an error is raised.
        2.  When a classical bit is used in a  measurement with an index higher than the number of classical bits an
            error is raised.
        """
        if self._number_of_clbits < 1:
            raise QiskitBackendError(f"Invalid number of classical bits ({self._number_of_clbits})!")

        if self.max_measurement_index >= self._number_of_clbits:
            raise QiskitBackendError(f"Number of classical bits ({self._number_of_clbits}) is not sufficient for "
                                     f"storing the outcomes of the experiment")

    def validate_unsupported_measurements(self) -> None:
        """ Validate unsupported measurements (for not full state projection algorithms)

        Certain measurements cannot be handled correctly because cQASM isn't as flexible as Qiskit in measuring to
        specific classical bits. Therefore some Qiskit constructions are not supported in QI:

            1. When a quantum register is measured to different classical registers
            2. When a classical register is used for the measurement of more than one quantum register

        :raises QiskitBackendError: When the circuit contains an invalid measurement
        """
        for q1, c1 in self._measurements_reg:
            for q2, c2 in self._measurements_reg:
                if q1 == q2 and c1 != c2:
                    raise QiskitBackendError(f'Measurement of qubit {q1} to different classical registers is not '
                                             f'supported')
                if q1 != q2 and c1 == c2:
                    raise QiskitBackendError(f'Measurement of different qubits to the same classical register {c1} '
                                             f'is not supported')

    def qubit_to_classical_hex(self, qubit_register: str) -> str:
        """ This function converts the qubit register data to the hexadecimal representation of the classical state.

        :param qubit_register: The measured value of the qubits represented as int.

        :return:
            The hexadecimal value of the classical state.
        """
        qubit_state = ('{0:0{1}b}'.format(int(qubit_register), self.nr_of_qubits))
        classical_state = ['0'] * self.nr_of_clbits
        for q, c in self._measurements_state:
            classical_state[c] = qubit_state[q]
        classical_state_str = ''.join(classical_state)
        classical_state_hex = hex(int(classical_state_str, 2))
        return classical_state_hex



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from io import StringIO
from typing import Optional, Tuple, List

import numpy as np

from qiskit.qobj import QasmQobjInstruction
from quantuminspire.exceptions import ApiError
from quantuminspire.qiskit.measurements import Measurements


class CircuitToString:
    """ Contains the translational elements to convert the Qiskit circuits to cQASM code."""

    def __init__(self, basis_gates: List[str], measurements: Measurements, full_state_projection: bool = False) -> None:
        """
        :param basis_gates: List of basis gates from the configuration.
        :param measurements: The measured qubits/classical bits and the number of qubits and classical bits.
        :param full_state_projection: When full_state_projection = True, no measurement statements are added.
        """
        self.basis_gates = basis_gates.copy()
        if len(self.basis_gates) > 0:
            self.basis_gates.append("measure")
            self.basis_gates.append("u")
        self.bfunc_instructions: List[QasmQobjInstruction] = []
        self.full_state_projection = full_state_projection
        self.measurements = measurements

    @staticmethod
    def _gate_not_supported(_stream: StringIO, instruction: QasmQobjInstruction, _binary_control: Optional[str] = None)\
            -> None:
        """ Called when a gate is not supported with the backend. Throws an exception (ApiError)

        :param instruction: The Qiskit instruction to translate to cQASM.

        :raises ApiError: the gate is not supported by the circuit parser.

        """
        if hasattr(instruction, 'conditional'):
            raise ApiError(f"Conditional gate 'c-{instruction.name.lower()}' not supported")

        raise ApiError(f"Gate '{instruction.name.lower()}' not supported")

    @staticmethod
    def _cz(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the controlled Z element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('CZ q[{0}], q[{1}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_cz(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled controlled Z element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-CZ {0}q[{1}], q[{2}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _cx(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the controlled X element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('CNOT q[{0}], q[{1}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_cx(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled controlled X element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-CNOT {0}q[{1}], q[{2}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _ccx(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Toffoli element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Toffoli q[{0}], q[{1}], q[{2}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_ccx(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary controlled Toffoli element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-Toffoli {0}q[{1}], q[{2}], q[{3}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _h(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the H element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('H q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_h(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled H element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-H {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _id(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the ID element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('I q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_id(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled ID element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-I {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _s(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the S element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('S q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_s(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled S element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-S {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _sdg(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Sdag element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Sdag q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_sdg(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Sdag element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-Sdag {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _swap(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the SWAP element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('SWAP q[{0}], q[{1}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_swap(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled SWAP element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-SWAP {0}q[{1}], q[{2}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _t(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the T element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('T q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_t(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled T element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-T {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _tdg(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Tdag element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Tdag q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_tdg(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Tdag element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-Tdag {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _x(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the X element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('X q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_x(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled X element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-X {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _y(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Y element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Y q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_y(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Y element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-Y {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _z(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Z element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Z q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_z(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Z element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        stream.write('C-Z {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _r(stream: StringIO, instruction: QasmQobjInstruction, axis: str) -> None:
        """ Translates the Rotation element for an axis (x,y,z).

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param axis: The axis for which the Rotation operator is parsed ('x', 'y' or 'z').

        """
        angle_q0 = float(instruction.params[0])
        stream.write('R{0} q[{1}], {2:.6f}\n'.format(axis, *instruction.qubits, angle_q0))

    @staticmethod
    def _c_r(stream: StringIO, instruction: QasmQobjInstruction, axis: str, binary_control: str) -> None:
        """ Translates the binary-controlled Rotation element for an axis (x,y,z).

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param axis: The axis for which the Rotation operator is parsed ('x', 'y' or 'z').
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        angle_q0 = float(instruction.params[0])
        stream.write('C-R{0} {1}q[{2}], {3:.6f}\n'.format(axis, binary_control, *instruction.qubits, angle_q0))

    @staticmethod
    def _rx(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Rx element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'x')

    @staticmethod
    def _c_rx(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Rx element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        CircuitToString._c_r(stream, instruction, 'x', binary_control)

    @staticmethod
    def _ry(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Ry element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'y')

    @staticmethod
    def _c_ry(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Ry element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        CircuitToString._c_r(stream, instruction, 'y', binary_control)

    @staticmethod
    def _rz(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the Rz element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'z')

    @staticmethod
    def _c_rz(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled Rz element.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        CircuitToString._c_r(stream, instruction, 'z', binary_control)

    @staticmethod
    def _u(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the U element to U3.

        The u element is used by Qiskit for the u_base gate and when a u0-gate
        is used in the circuit but not supported as a basis gate for the backend.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._u3(stream, instruction)

    @staticmethod
    def _c_u(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled U element to binary-controlled U3.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        CircuitToString._c_u3(stream, instruction, binary_control)

    @staticmethod
    def _u1(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the U1(lambda) element to U3(0, 0, lambda).

        A copy of the circuit is made to prevent
        side effects for the caller.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params[0:0] = (0, 0)
        CircuitToString._u3(stream, temp_instruction)

    @staticmethod
    def _c_u1(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled U1(lambda) element to U3(0, 0, lambda).

        A copy of the circuit is
        made to prevent side effects for the caller.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params[0:0] = (0, 0)
        CircuitToString._c_u3(stream, temp_instruction, binary_control)

    @staticmethod
    def _p(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the p element to u1.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._u1(stream, instruction)

    @staticmethod
    def _c_p(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the c-p element to c-u1.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        CircuitToString._c_u1(stream, instruction, binary_control)

    @staticmethod
    def _u3(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the U3(theta, phi, lambda) element to 3 rotation gates.

        Any single qubit operation (a 2x2 unitary matrix) can be written as the product of rotations.
        As an example, a unitary single-qubit gate can be expressed as a combination of
        Rz and Ry rotations (Nielsen and Chuang, 10th edition, section 4.2).
        U(theta, phi, lambda) = Rz(phi)Ry(theta)Rz(lambda).
        Note: The expression above is the matrix multiplication, when implementing this in a gate circuit,
        the gates need to be executed in reversed order.
        Any rotation of 0 radials is left out of the resulting circuit.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        gates = ['Rz', 'Ry', 'Rz']
        angles = list(float(instruction.params[i]) for i in [2, 0, 1])
        index_q0 = [instruction.qubits[0]] * 3
        for triplet in zip(gates, index_q0, angles):
            if triplet[2] != 0:
                stream.write('{0} q[{1}], {2:.6f}\n'.format(*triplet))

    @staticmethod
    def _c_u3(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled U3(theta, phi, lambda) element to 3 rotation gates.

        See gate :meth:`~._u3` for more information.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """
        gates = ['C-Rz', 'C-Ry', 'C-Rz']
        binary_controls = [binary_control] * 3
        angles = list(float(instruction.params[i]) for i in [2, 0, 1])
        index_q0 = [instruction.qubits[0]] * 3
        for quadruplets in zip(gates, binary_controls, index_q0, angles):
            if quadruplets[3] != 0:
                stream.write('{0} {1}q[{2}], {3:.6f}\n'.format(*quadruplets))

    @staticmethod
    def _barrier(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the | element for a variable number of qubits.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('barrier q[{0}]\n'.format(','.join(map(str, instruction.qubits))))

    @staticmethod
    def _c_barrier(stream: StringIO, instruction: QasmQobjInstruction, binary_control: str) -> None:
        """ Translates the binary-controlled | element. No cQASM is added for this gate.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        :param binary_control: The multi-bits control string. The gate is executed when all specified classical bits
        are 1.

        """

    @staticmethod
    def _delay(stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the delay element for a qubit. In cQASM wait parameter is int and the unit is hardware cycles.
        Only the Qiskit default unit "dt" will work correctly with cQASM, i.e. integer time unit depending on the
        target backend.

        In qiskit/circuit/delay.py multiple units are defined for delay instruction.
        In qiskit/circuit/instruction.py assemble() method the unit of the delay instruction is not passed. Only the
        parameter (which is the value of the delay instruction) is taken.
        Here we cannot convert delays originally in another unit than dt to dt, which is the unit for wait in cQASM.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        wait_period = int(instruction.params[0])
        stream.write('wait q[{0}], {1}\n'.format(*instruction.qubits, wait_period))

    def _measure(self, stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Translates the measure element. No cQASM is added for this gate when FSP is used.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        if not self.full_state_projection:
            stream.write('measure q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def get_mask_data(mask: int) -> Tuple[int, int]:
        """ Get mask data

        A mask is a continuous set of 1-bits with a certain length. This method returns the lowest bit of
        the mask and the length of the mask.

        Examples:

        ============ ====================================
        ``76543210`` bit_nr
        ============ ====================================
        ``00111000`` lowest mask bit = 3, mask_length = 3
        ``00000001`` lowest mask bit = 0, mask_length = 1
        ``11111111`` lowest mask bit = 0, mask_length = 8
        ``10000000`` lowest mask bit = 7, mask_length = 1
        ============ ====================================

        :param mask: The mask to get the mask data from.

        :return: The mask data, i.e. a tuple (lowest_bit_number, mask_length)
        """
        # Precondition: mask != 0
        if mask == 0:
            return -1, 0
        mask_length = 0
        bit_value = 1
        bit_nr = 0
        while not mask & bit_value:
            bit_value <<= 1
            bit_nr += 1
        lowest_mask_bit = bit_nr
        while mask & bit_value:
            mask_length += 1
            bit_value <<= 1
        return lowest_mask_bit, mask_length

    def _parse_bin_ctrl_gate(self, stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Parses a binary controlled gate.

        A binary controlled gate name is preceded by 'c-'.
        The gate is executed when a specific measurement is true. Multiple measurement outcomes are used
        to control the quantum operation. This measurement is a combination of classical bits being 1 and others
        being 0. Because cQASM only supports measurement outcomes of 1, any other bits in the
        masked bit pattern first have to be inverted with the not-operator. The same inversion also has to
        take place after the binary controlled quantum operation.
        The mask can be one or more bits and start at any bit depending on the instruction and the declaration
        of classical bits.
        The resulting stream will be expanded with something like:
        not b[the 0-bits in the value relative to the mask changed to 1]
        c-gate [classical bits in the mask], other arguments
        not b[the 0-bits reset to 0 again]
        When the c-gate results in an empty string (e.g. binary controlled u(0, 0, 0) or barrier gate),
        nothing is added to the stream.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.

        """
        conditional_reg_idx = instruction.conditional
        conditional = next((x for x in self.bfunc_instructions if x.register == conditional_reg_idx), None)
        if conditional is None:
            raise ApiError(f'Conditional not found: reg_idx = {conditional_reg_idx}')
        self.bfunc_instructions.remove(conditional)

        conditional_type = conditional.relation
        if conditional_type != '==':
            raise ApiError(f'Conditional statement with relation {conditional_type} not supported')
        mask = int(conditional.mask, 16)
        if mask == 0:
            raise ApiError(f'Conditional statement {instruction.name.lower()} without a mask')
        lowest_mask_bit, mask_length = self.get_mask_data(mask)
        val = int(conditional.val, 16)
        masked_val = mask & val

        # form the negation to the 0-values of the measurement registers, when value == mask no bits are negated
        negate_zeroes_line = ''
        if masked_val != mask:
            negate_zeroes_line = 'not b[' + ','.join(
                str(self.measurements.get_qreg_for_conditional_creg(i)) for i in
                range(lowest_mask_bit, lowest_mask_bit + mask_length) if not (masked_val & (1 << i))) + ']\n'

        if mask_length == 1:
            binary_control = f'b[{self.measurements.get_qreg_for_conditional_creg(lowest_mask_bit)}], '
        else:
            # form multi bits control - qasm-single-gate-multiple-qubits
            binary_control = f'b[' + ','.join(str(self.measurements.get_qreg_for_conditional_creg(i)) for i in
                                              range(lowest_mask_bit, lowest_mask_bit + mask_length)) + '], '

        with StringIO() as gate_stream:
            # add the gate
            gate_name = f'_c_{instruction.name.lower()}'
            gate_function = getattr(self, gate_name, getattr(self, "_gate_not_supported"))
            gate_function(gate_stream, instruction, binary_control)
            line = gate_stream.getvalue()
            if len(line) != 0:
                # negate the measurement registers that has to be 0
                stream.write(negate_zeroes_line)
                stream.write(line)
                # reverse the measurement registers that had to be 0
                stream.write(negate_zeroes_line)

    def parse(self, stream: StringIO, instruction: QasmQobjInstruction) -> None:
        """ Parses a gate.

        For each type of gate a separate (private) parsing method is defined and called.
        The resulting cQASM code is written to the stream. When the gate is a binary controlled
        gate, Qiskit uses two instructions to handle it. The first instruction is a so-called bfunc with the
        conditional information (mask, value to check etc.) which is stored for later use.
        The next instruction is the actual gate which must be executed conditionally. The parsing is
        forwarded to method _parse_bin_ctrl_gate which reads the earlier stored bfunc.
        When a gate is not supported _gate_not_supported is called which raises an exception.

        :param stream: The string-io stream to where the resulting cQASM is written.
        :param instruction: The Qiskit instruction to translate to cQASM.
        """
        if instruction.name == 'bfunc':
            self.bfunc_instructions.append(instruction)
        elif len(self.basis_gates) > 0 and instruction.name.lower() not in self.basis_gates:
            self._gate_not_supported(stream, instruction)
        elif hasattr(instruction, 'conditional'):
            self._parse_bin_ctrl_gate(stream, instruction)
        else:
            gate_name = f'_{instruction.name.lower()}'
            gate_function = getattr(self, gate_name, getattr(self, "_gate_not_supported"))
            gate_function(stream, instruction)



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from typing import List, Optional, Any, Dict, Callable, TYPE_CHECKING

from qiskit.providers import JobError, JobTimeoutError, JobV1 as Job
from qiskit.providers.jobstatus import JobStatus, JOB_FINAL_STATES
from qiskit.result.models import ExperimentResult
from qiskit.qobj import QasmQobj, QasmQobjExperiment
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.job import QuantumInspireJob
from quantuminspire.qiskit.qi_result import QIResult
from quantuminspire.version import __version__ as quantum_inspire_version

if TYPE_CHECKING:
    from backend_qx import QuantumInspireBackend


class QIJob(Job):  # type: ignore
    """
    A Qiskit job that is executed on the Quantum Inspire platform. A QIJob is normally created by calling
    ``execute`` (which triggers a `run` on the QuantumInspireBackend ``qi_backend``).

    .. code::

        qc = QuantumCircuit(5, 5)
        qc.h(0)
        qc.cx(0, range(1, 5))
        qc.measure_all()

        job = execute(qc, qi_backend, shots=1024)
        result = job.result()

    The return value of ``execute`` is an instance of QIJob (Qiskit job) and is the equivalent to the Quantum
    Inspire project. It is a container to handle one or more (asynchronous) Qiskit circuits or experiments.
    A Qiskit circuit is equivalent to a Quantum Inspire job.

    To get the Quantum Inspire project-id from ``job`` do:

    .. code::

        qi_project_id = job.job_id()

    To get the list of Quantum Inspire jobs (in dictionary format) from ``job`` do:

    .. code::

        qi_jobs = job.get_jobs()

    QIJob can also be recreated using a ``job_id``, being a Quantum Inspire project-id:

    .. code::

        qi_backend = QI.get_backend('QX single-node simulator')
        job = qi_backend.retrieve_job(job_id)

    """

    def __init__(self, backend: QuantumInspireBackend, job_id: str, api: QuantumInspireAPI,
                 qobj: Optional[QasmQobj] = None) -> None:
        """
        A QIJob object is normally not constructed directly.

        :param backend: A Quantum Inspire backend.
        :param job_id: Id of the job as provided by the Quantum Inspire API.
        :param api: A Quantum Inspire API.
        :param qobj: A Qiskit quantum object.
        """
        self._api: QuantumInspireAPI = api
        super().__init__(backend, job_id)
        self.experiments: Optional[List[QasmQobjExperiment]] = None
        self.jobs: List[QuantumInspireJob] = []
        self._status: JobStatus = JobStatus.INITIALIZING
        self._qobj: Optional[QasmQobj] = qobj
        if self._qobj is not None:
            self._job_id = ''  # invalidate _job_id
        else:
            self.status()

    def set_job_id(self, job_id: str) -> None:
        """ Overwrite the job_id with a new id.

        :param job_id: New id for the QIJob. Used in the use case for linking the job to the user-given QI project that
            must contain the job.

        """
        self._job_id = job_id

    def submit(self) -> None:
        """
        Submit a job to the Quantum Inspire platform.

        :raises JobError: An error if the job has already been submitted.
        """
        if self._job_id:
            raise JobError('Job has already been submitted!')
        self._job_id = self._backend.run(circuits=self._qobj)

    def _result(self, result_function: Callable[[Job], List[ExperimentResult]], timeout: Optional[float] = None,
                wait: float = 0.5) -> QIResult:
        """ Return the result for the experiments.

        :param result_function: backend function for fetching the requested results.
        :param timeout: Timeout in seconds.
        :param wait: Wait time between queries to the Quantum Inspire platform.

        :return:
            Result object containing results from the experiments.

        :raises JobTimeoutError: If timeout is reached.
        :raises QiskitBackendError: If an error occurs during simulation.
        """
        start_time = time.time()
        while self.status() not in JOB_FINAL_STATES:
            elapsed_time = time.time() - start_time
            if timeout is not None and elapsed_time > timeout:
                raise JobTimeoutError('Failed getting result: timeout reached.')
            time.sleep(wait)

        experiment_results = result_function(self)
        total_time_taken = sum(getattr(experiment_result, "time_taken", 0.0) for
                               experiment_result in experiment_results)

        return QIResult(backend_name=self._backend.backend_name, backend_version=quantum_inspire_version,
                        job_id=self.job_id(), qobj_id=self.job_id(), success=True, results=experiment_results,
                        status=self.status().name, time_taken=total_time_taken)

    def result(self, timeout: Optional[float] = None, wait: float = 0.5) -> QIResult:
        """ Return the result for the experiments in the latest run for this project.

        :param timeout: Timeout in seconds.
        :param wait: Wait time between queries to the Quantum Inspire platform.

        :return:
            QIResult object containing the result.

        :raises JobTimeoutError: If timeout is reached.
        :raises QiskitBackendError: If an error occurs during simulation.
        """
        return self._result(self._backend.get_experiment_results_from_latest_run, timeout, wait)

    def result_all_jobs(self, timeout: Optional[float] = None, wait: float = 0.5) -> QIResult:
        """ Return the result for the experiments for all the existing jobs in the project.

        :param timeout: Timeout in seconds.
        :param wait: Wait time between queries to the Quantum Inspire platform.

        :return:
            QIResult object containing the result.

        :raises JobTimeoutError: If timeout is reached.
        :raises QiskitBackendError: If an error occurs during simulation.
        """
        return self._result(self._backend.get_experiment_results_from_all_jobs, timeout, wait)

    def cancel(self) -> None:
        """ Cancel the job and delete the project. """
        self._api.delete_project(int(self._job_id))

    def get_jobs(self) -> List[Dict[str, Any]]:
        """ Gets the Quantum Inspire jobs that were submitted in the latest run. These job were added with add_job.

        :return:
            List of jobs with their properties for the jobs that were added for the experiments submitted when running
            this instance of QIJob.
            An empty list is returned when there were no jobs added.

        :raises ApiError: If the job for the job identified does not exist.
        """
        ret = [self._api.get_job(job.get_job_identifier()) for job in self.jobs]
        return ret

    def add_job(self, job: QuantumInspireJob) -> None:
        """ Add a Quantum Inspire job to the list. The list contains the (Quantum Inspire) jobs created for the
        submitted experiments in this particular QIJob.

        :param job: QuatumInspireJob (submitted) that has to be added to the list of jobs created for the experiments
         in QIJob.
        """
        self.jobs.append(job)

    def queue_position(self, refresh: bool = False) -> Optional[int]:
        """
        Return the position for this Job in the Quantum Inspire queue (when in status QUEUED).
        Currently we don't have this info available.

        :param refresh: If ``True``, re-query the server to get the latest value.
                Otherwise return the cached value, when available. Not used.

        :return:
            The queue position of the job. Currently None (not available).
        """
        return None

    def status(self) -> JobStatus:
        """
        Check the status of the jobs submitted in the latest run.

        :return:
            The status of the job.
        """
        jobs = self.get_jobs()
        number_of_jobs = len(jobs)
        cancelled = len([job for job in jobs if job['status'] == 'CANCELLED'])
        running = len([job for job in jobs if job['status'] == 'RUNNING'])
        completed = len([job for job in jobs if job['status'] == 'COMPLETE'])

        if number_of_jobs == 0:
            self._status = JobStatus.INITIALIZING
        elif 0 < cancelled < number_of_jobs:
            self._status = JobStatus.ERROR
        elif cancelled == number_of_jobs:
            self._status = JobStatus.CANCELLED
        elif running > 0 or (0 < completed < number_of_jobs):
            self._status = JobStatus.RUNNING
        elif completed == number_of_jobs:
            self._status = JobStatus.DONE
        else:
            self._status = JobStatus.QUEUED
        return self._status



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module api
==========

.. autoclass:: QuantumInspireAPI
   :members:
"""

import re
import itertools
import logging
import time
import uuid
from typing import Type, List, Dict, Union, Optional, Any, Tuple
from urllib.parse import urljoin

from coreapi.auth import TokenAuthentication, AuthBase
from coreapi.client import Client
from coreapi.exceptions import CoreAPIException, ErrorMessage
from coreapi.transports import HTTPTransport

from quantuminspire.credentials import load_account
from quantuminspire.exceptions import ApiError, AuthenticationError
from quantuminspire.job import QuantumInspireJob

logger = logging.getLogger(__name__)
V1_MEASUREMENT_BLOCK_INDEX = -1  # -1 for last block
QI_URL = 'https://api.quantum-inspire.com'


class VersionedAPITransport(HTTPTransport):  # type: ignore[misc]
    """ VersionedAPITransport makes it possible to address a specific version of the API of quantum inspire.
    """
    def __init__(self, api_version: str = '2.0', auth: AuthBase = None) -> None:
        self._api_version = api_version
        super().__init__(auth=auth, headers=self.headers)

    @property
    def headers(self) -> Dict[str, str]:
        """ Return accept string for versioned transport """
        return {
            'accept': f'application/coreapi+json, application/vnd.coreapi+json, */*; version={self._api_version};'
        }


class QuantumInspireAPI:
    def __init__(self, base_uri: str = QI_URL, authentication: Optional[AuthBase] = None,
                 project_name: Optional[str] = None,
                 coreapi_client_class: Type[Client] = Client) -> None:
        """ Python interface to the Quantum Inspire API (Application Programmer Interface).

        The Quantum Inspire API supplies an interface for executing programs and can be used to access the
        different entities in the Quantum Inspire database needed for running the programs.
        The entities for which an interface is provided are:

            * Backend types: Depending on the user account more qubits can be used and simulated on faster hardware.
            * Projects: Executing programs is done from a project. Projects keep track of the other entities.
            * Jobs: A job contains all the information and the parameters needed for executing the program.
            * Assets: A container for a cQASM program. Is part of a job.
            * Results: After the job is finished, the results are gathered in the result-entity.

        QuantumInspireAPI is a convenient interface (or wrapper) to the low level API and hides details for
        requesting data (via get) and performing operations on the different entities (via actions).

        The API version that is used by SDK version 2.0.0 and higher defaults to 2.0.
        Version 2.0 is introduced for serving backends that support multiple measurement algorithms.
        Backends that support multiple measurement algorithms have a flag defined called 'multiple_measurement' in their
        backend type structure (See :meth:`~.get_default_backend_type`).

        To support multiple measurement the results structure has changed to hold the results for each measurement block
        defined in a multiple measurement algorithm. Backends that do not support multiple measurement algorithms have
        1 measurement block. See :meth:`~.get_result` for details.

        For more documentation:
            * The knowledge base on: https://www.quantum-inspire.com/kbase/low-level-api/
            * The REST API can be found on: https://api.quantum-inspire.com/
            * The Core API schema is published on: https://api.quantum-inspire.com/schema/

        :param base_uri: The base uri of the Quantum Inspire API-location where the schema can be found
                in path 'schema/').
        :param authentication: The authentication, can be one of the following :mod:`coreapi` authentications:

                 * :class:`~coreapi.auth.BasicAuthentication(email, password)`,
                   HTTP authentication with valid email/password.
                 * :class:`~coreapi.auth.TokenAuthentication(token, scheme="token")`,
                   token authentication with a valid API-token.

                When authentication is :obj:`None`, a token is read from the default location.
        :param project_name: The name of the project used for executing the jobs.
        :param coreapi_client_class: Coreapi client to interact with the API through a schema.
                Default set to :class:`coreapi.Client`.

        :raises AuthenticationError: When no authentication is given
                and the token could not be loaded from the default location.
        :raises ApiError: When the schema could not be loaded.

        .. note::
            When no project name is given, a temporary project is created for the job and deleted after the job
            has finished. When a project name is given, a project is created if it does not exist, but re-used
            if a project with that name already exists. In either case, the project will not be deleted when
            the job has finished.

        """
        if authentication is None:
            token = load_account()
            if token is not None:
                authentication = TokenAuthentication(token, scheme="token")
            else:
                raise AuthenticationError('No credentials have been provided or found on disk')

        self.__client = coreapi_client_class(transports=[VersionedAPITransport(auth=authentication)])
        self.project_name = project_name
        self.base_uri = base_uri if base_uri[-1] == '/' else base_uri + '/'
        self.enable_fsp_warning = True
        try:
            self._load_schema()
        except (CoreAPIException, TypeError) as ex:
            raise ApiError(f'Could not connect to {base_uri}') from ex

    def _get(self, uri_path: str) -> Any:
        """ Method for making requests to the coreapi client instance to get some piece of information.

        The information requested depends on the uri_path parameter.

        :param uri_path: The URL where to request the data.

        :raises TypeError: When the uri_path is not correct.
        :raises CoreAPIException: When the get was not successful. Possible causes are:

            * The schema is not loaded, or
            * A network error occurred.

        :raises ErrorMessage: When the result of the get is not successful.

        :return: The resulting data from the get-request. The structure of the data depends on the request.

        .. note:
            This method will be made private in the near future. Usage is discouraged.

        """
        return self.__client.get(uri_path)

    def show_fsp_warning(self, enable: bool = True) -> None:
        """ Enable or disable the warning that is printed when a non-FSP (full state projection) job is about to run.

        :param enable: when True the fsp-warning is shown, otherwise not.

        """
        self.enable_fsp_warning = enable

    def _action(self, action: List[str], params: Optional[Dict[str, Any]] = None) -> Any:
        """ Adapter for performing an action on an object via the Quantum Inspire API.

        :param action: Path in the schema hierarchy selecting the requested action.
        :param params: Some actions may accept a set of parameters with names as keys.

        :raises CoreAPIException: When the action was not successful. Possible causes are:

            * One of the actions or parameters not supported,
            * the schema is not loaded, or
            * a network error occurred.

        :raises CoreAPIException.ErrorMessage: When the result of the action is not successful.

        :return: The resulting data from the action-request. The structure of the data depends on the request.
            Can be None when there is no content in the response.
        """
        return self.__client.action(self.document, action, params=params)

    def _load_schema(self) -> None:
        """ Loads the schema with metadata that explains how the api-data is structured."""
        self.document = self._get(urljoin(self.base_uri, 'schema/'))

    def list_backend_types(self) -> None:
        """ Prints the backend types with the name and the maximum number of qubits it supports."""
        backends = self.get_backend_types()
        for backend in backends:
            print(f'Backend type: {backend["name"]}, number of qubits: {backend["number_of_qubits"]}')

    def get_default_backend_type(self) -> Dict[str, Any]:
        """ Gets the properties of the default backend type.

        :return: The default backend type with all of its properties:

            =================================== =========== ============================================================
            Key                                 Type        Description
            =================================== =========== ============================================================
            ``url``                             str         The url for the backend type.
            ``name``                            str         Name of the backend.
            ``is_hardware_backend``             bool        Indicates whether the backend is a hardware backend (True)
                                                            or a simulating backend (False).
            ``required_permission``             str         Describes the permission that is required
                                                            to use this backend.
            ``number_of_qubits``                int         Maximum number of qubits the backend supports.
            ``description``                     str         Short description of the backend.
            ``topology``                        dict        Dictionary with property `edges` that contains a list of
                                                            tuples that define qubit connectivity for 2-qubit gates.
            ``is_allowed``                      bool        Indicates whether the user is allowed to use this backend.
            ``status``                          str         Status of the backend.
            ``status_message``                  str         Extra info about the status of the backend.
            ``chip_image_id``                   str         Unique identification of the chip.
            ``calibration``                     str         Calibration information (url).
            ``allowed_operations``              dict        The gates/operations names that the backend can handle.
            ``default_number_of_shots``         int         The default number of shots for an experiment.
            ``max_number_of_shots``             int         The maximum number of shots for an experiment.
            ``max_number_of_simultaneous_jobs`` int         The maximum number of jobs that is allowed to be queued for
                                                            the backend simultaneously (0 = no limit).
            ``operations_count``                dict        The maximum number of gates that is allowed in an experiment
                                                            for each qubit separately and for the experiment in total
                                                            (0 = no limit).
            ``flags``                           list[str]   flags defining capabilities of the backends. E.g.
                                                            'multiple_measurement', 'parallel_computing'
            =================================== =========== ============================================================

        """
        return dict(self._action(['backendtypes', 'default', 'list']))

    def get_backend_types(self) -> List[Dict[str, Any]]:
        """ Gets a list of backend types with properties.

        :return:
            A list of backend types with all of their properties.
            See :meth:`~.get_default_backend_type`
            for a description of the backend properties.
        """
        ret: List[Dict[str, Any]] = self._action(['backendtypes', 'list'])
        return ret

    def get_backend_type_by_id(self, backend_type_id: int) -> Dict[str, Any]:
        """ Gets the properties of a specific backend type, given the backend type id.

        :param backend_type_id: The backend identification number.

        :raises ApiError: When the backend type indicated by backend_type_id does not exist.

        :return:
            The requested backend type indicated by backend_type_id with all of its properties.
            See :meth:`~.get_default_backend_type`
            for a description of the backend type properties.
        """
        try:
            backend_type = self._action(['backendtypes', 'read'], params={'id': backend_type_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Backend type with id {backend_type_id} does not exist!') from err_msg
        return dict(backend_type)

    def get_backend_type_by_name(self, backend_name: str) -> Dict[str, Any]:
        """ Gets the properties of a backend type, given the backend name (case insensitive).

        :param backend_name: The backend name.

        :raises ApiError: When the backend name does not exist.

        :return:
            The properties of the backend type of the specific backend.
            See :meth:`~.get_default_backend_type`
            for a description of the backend type properties.
        """
        backend_type = next((backend for backend in self.get_backend_types()
                            if backend['name'].lower() == backend_name.lower()), None)
        if backend_type is None:
            raise ApiError(f'Backend type with name {backend_name} does not exist!')
        return dict(backend_type)

    def get_backend_type(self, identifier: Optional[Union[int, str]] = None) -> Dict[str, Any]:
        """Gets the properties of the backend type indicated by `identifier`.

        If no `identifier` is given,
        the default backend type will be returned. With an `identifier`
        of type :class:`string` or :class:`int`,
        the backend type will be searched by name or id respectively.

        :param identifier: The backend type identifier.

        :raises ApiError: If the requested backend type does not exist.
        :raises ValueError: If the backend type identifier is not of the correct type.

        :return:
            This method returns the default backend type or the backend type as identified by
            the `identifier` parameter , or the default backend type if no `identifier` is provided.
            See :meth:`~.get_default_backend_type`
            for a description of the backend type properties.
        """
        if identifier is None:
            return self.get_default_backend_type()
        if isinstance(identifier, int):
            return self.get_backend_type_by_id(identifier)
        if isinstance(identifier, str):
            return self.get_backend_type_by_name(identifier)
        raise ValueError('Identifier should be of type int, str or None!')

    #  projects  #

    def list_projects(self) -> None:
        """ Prints a list of projects.

        Prints all the projects registered to the current user the API is authenticated for.
        For each project the name, id and backend type is printed.
        """
        projects = self.get_projects()
        for project in projects:
            print(f'Project name: {project["name"]}, id: {project["id"]}, backend type: {project["backend_type"]}')

    def get_project(self, project_id: int) -> Dict[str, Any]:
        """ Gets the properties of a project, given the project id.

        :param project_id: The project identification number.

        :raises ApiError: If the requested project does not exist.

        :return: The properties describing the project:

            =========================== =========== ==============================================================
            Key                         Type        Description
            =========================== =========== ==============================================================
            ``url``                     str         The url for this project.
            ``id``                      int         Unique id of the project.
            ``name``                    str         Name of the project.
            ``owner``                   int         Url to get the owner of the project.
            ``assets``                  str         Url to get the assets of the project.
            ``backend_type``            str         Url to get the backend type of the project.
            ``default_number_of_shots`` int         Default number of executions for this project.
            ``created``                 str         Date/time when the project was created.
            ``number_of_jobs``          int         Number of jobs managed within this project.
            ``last_saved``              str         Date/time when the project was saved.
            =========================== =========== ==============================================================

        """
        try:
            project = self._action(['projects', 'read'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Project with id {project_id} does not exist!') from err_msg
        return dict(project)

    def get_projects(self) -> List[Dict[str, Any]]:
        """ Gets all the projects registered to the user the API is currently authenticated for.

        :return:
            The projects with all of their properties.
            See :meth:`~.get_project` for a description of the project properties.
        """
        ret: List[Dict[str, Any]] = self._action(['projects', 'list'])
        return ret

    def create_project(self, name: str, default_number_of_shots: int, backend_type: Dict[str, Any]) -> Dict[str, Any]:
        """ Creates a new project for executing jobs.

        :param name: The name for the project. The name need not be unique.
        :param default_number_of_shots: The default number of executions of the program before collecting
            the results.
        :param backend_type: The properties of the backend type.

        :return:
            The newly created project with all of its properties.
            See :meth:`~.get_project` for a description of the project properties.
        """
        payload = {
            'name': name,
            'default_number_of_shots': default_number_of_shots,
            'backend_type': backend_type['url'],
        }
        return dict((self._action(['projects', 'create'], params=payload)))

    def delete_project(self, project_id: int) -> None:
        """ Delete a project.

        Deletes the project identified by `project_id` together with all its assets, jobs and results.
        Only projects can be deleted that are registered for the user the API is currently authenticated for.

        :param project_id: The project identification number.

        :raises ApiError: If the project identified by project_id does not exist.
        """
        payload = {
            'id': project_id
        }
        try:
            self._action(['projects', 'delete'], params=payload)
        except ErrorMessage as err_msg:
            raise ApiError(f'Project with id {project_id} does not exist!') from err_msg

    #  jobs  #

    def list_jobs(self) -> None:
        """ Prints a list of jobs.

        Prints a list of all the jobs registered to the current user the API is authenticated for.
        For each job the name, identification number and status is printed.
        """
        jobs = self.get_jobs()
        for job in jobs:
            print(f'Job name: {job["name"]}, id: {job["id"]}, status: {job["status"]}')

    def get_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the properties of a job, given the job id.

        :param job_id: The job identification number.

        :raises ApiError: If the requested job does not exist.

        :return: The properties describing the job:

            ========================= =========== ==================================================================
            Key                       Type        Description
            ========================= =========== ==================================================================
            ``url``                   str         The url for the job.
            ``name``                  str         Name of the circuit that is executed by this job.
            ``id``                    int         Unique id of the job.
            ``status``                str         Execution status of the job:
                                                  e.g. 'NEW', 'COMPLETE', 'CANCELLED', 'RUNNING'.
            ``input``                 str         Url to get the assets of the job.
            ``backend``               str         Url to get the backend the job is executed on.
            ``backend_type``          str         Url to get the backend type of the backend
                                                  the job is executed on.
            ``results``               str         Url to get the results for the job.
            ``queued_at``             str         The date-time the job is queued at.
                                                  The format is 'yyyy-MM-ddTHH:mm:ss.SSSSSSZ' Zulu Time.
            ``number_of_shots``       int         Number of executions for this job.
            ``full_state_projection`` bool        Indicates if the backend uses full state projection
                                                  to determine the quantum state.
                                                  Used for optimizing simulations. For more information see:
                                                  https://www.quantum-inspire.com/kbase/optimization-of-simulations/
            ``user_data``             str         The user configuration data.
            ========================= =========== ==================================================================

        """
        try:
            job = self._action(['jobs', 'read'], params={'id': job_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Job with id {job_id} does not exist!') from err_msg
        return dict(job)

    def get_jobs(self) -> List[Dict[str, Any]]:
        """ Gets all the jobs registered to projects for the user the API is currently authenticated for.

        :return:
            The jobs with all of its properties.
            See :meth:`~.get_job` for a description of the job properties.
        """
        ret: List[Dict[str, Any]] = self._action(['jobs', 'list'])
        return ret

    def get_jobs_from_asset(self, asset_id: int) -> List[Dict[str, Any]]:
        """ Gets the jobs with its properties for an asset, given the asset id.

        :param asset_id: The asset identification number.

        :return:
            List of jobs with its properties for the asset with identification `asset_id`.
            An empty list is returned when the asset has no jobs.

        :raises ApiError: If the asset identified by `asset_id` does not exist.
        """
        try:
            jobs = self._action(['assets', 'jobs', 'list'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Asset with id {asset_id} does not exist!') from err_msg
        ret: List[Dict[str, Any]] = jobs
        return ret

    def get_jobs_from_project(self, project_id: int) -> List[Dict[str, Any]]:
        """ Gets the jobs with its properties for a single project, given the project id.

        :param project_id: The project identification number.

        :return:
            List of jobs with its properties for the project with identification `project_id`.
            An empty list is returned when the project has no jobs.

        :raises ApiError: If the project identified by `project_id` does not exist.
        """
        try:
            jobs = self._action(['projects', 'jobs', 'list'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Project with id {project_id} does not exist!') from err_msg
        ret: List[Dict[str, Any]] = jobs
        return ret

    def delete_job(self, job_id: int) -> None:
        """ Deletes the job identified by `job_id`.

        Only jobs can be deleted that are registered for the user the API is currently authenticated for.

        :param job_id: The job identification number.

        :raises ApiError: If the job identified by `job_id` does not exist.
        """
        try:
            self._action(['jobs', 'delete'], params={'id': job_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Job with id {job_id} does not exist!') from err_msg

    def _create_job(self, name: str, asset: Dict[str, Any], number_of_shots: int,
                    backend_type: Dict[str, Any], full_state_projection: bool = False,
                    user_data: str = '') -> Dict[str, Any]:
        """ Creates a new job for executing cQASM code.

        This method is used by execute_qasm_async and indirectly by execute_qasm.

        :param name: The name for the job.
        :param asset: The asset with the cQASM code.
        :param number_of_shots: The number of executions before returning the result.
        :param backend_type: The backend type.
        :param full_state_projection: Used for optimizing simulations. For more information see:
            https://www.quantum-inspire.com/kbase/optimization-of-simulations/
        :param user_data: Data that the user wants to pass along with the job.

        :return:
            The properties describing the new job.
            See :meth:`~.get_job` for a description of the job properties.
        """
        full_state_projection = full_state_projection and not backend_type.get("is_hardware_backend", True)
        payload = {
            'status': 'NEW',
            'name': name,
            'input': asset['url'],
            'backend_type': backend_type['url'],
            'number_of_shots': number_of_shots,
            'full_state_projection': full_state_projection,
            'user_data': user_data
        }
        if not full_state_projection and self.enable_fsp_warning and not backend_type.get("is_hardware_backend", False):
            logger.warning("Your experiment can not be optimized and may take longer to execute, "
                           "see https://www.quantum-inspire.com/kbase/optimization-of-simulations/ for details.")
        try:
            return dict(self._action(['jobs', 'create'], params=payload))
        except (CoreAPIException, TypeError, ValueError) as err_msg:
            raise ApiError(f'Job with name {name} not created: {err_msg}') from err_msg

    #  results  #

    def list_results(self) -> None:
        """ Prints a list of results.

        Prints a list of all the results registered to the current user the API is authenticated for.
        For each result the identification number and creation date is printed.
        """
        results = self.get_results()
        for result in results:
            print(f'Result id: {result["id"]} (date: {result["created_at"]})')

    def get_result(self, result_id: int) -> Dict[str, Any]:
        """ Gets the results of the executed job, given the result_id.

        Using version 2.0 of the API, the results structure has changed to allow multiple measurement algorithms.
        Some items in the results structure (histogram, raw data, measurement mask, quantum states,
        measurement register) are defined as lists instead of single values. The lists hold the single values for each
        measurement block in the algorithm for backends that support multiple measurement. Backends that do not support
        multiple measurement always have a single measurement block and therefore return a single item in the
        results lists.
        For some items (raw data, measurement mask) in the results structure the type have changed to better express
        measured and not measured qubits. See :meth:`~.get_raw_data_from_result` for details.

        :param result_id: The result identification number.

        :raises ApiError: If the requested result does not exist.

        :return:
            The properties describing the result:

            ============================= ================ =============================================================
            Key                           Type             Description
            ============================= ================ =============================================================
            ``id``                        int              Unique id of the result.
            ``url``                       str              The url to get this result.
            ``job``                       str              The url to get the job that generated the result.
            ``created_at``                str              The date-time the result is created at.
                                                           The format is 'yyyy-MM-ddTHH:mm:ss.SSSSSSZ' Zulu Time.
            ``number_of_qubits``          int              Number of qubits in the circuit for this experiment.
            ``execution_time_in_seconds`` float            The execution time of the job.
            ``raw_text``                  str              Text string filled when an error occurred, else empty.
            ``raw_data_url``              str              Url to get the raw data of the result. The raw data exists of
                                                           a list of integer values depicting the state for each shot.
            ``histogram``                 list(dict)       The histograms as a list of dicts with tuples with state
                                                           (str) and its probability (float).
            ``histogram_url``             str              Url to get the histogram with probabilities. This results in
                                                           the dict as found in property ``histogram`` of result.
            ``measurement_mask``          list(list(0|1))  For each measurement block a list of 0 (not measured) or 1
                                                           (measured) values for all the qubits in the cQASM program.
            ``quantum_states_url``        str              Url to get a list of quantum states for each measurement
                                                           block.
            ``measurement_register_url``  str              Url to get a list of measurement registers for each
                                                           measurement block.
            ``calibration``               str              Url to get calibration information.
            ============================= ================ =============================================================
        """
        try:
            result = self._action(['results', 'read'], params={'id': result_id, })
        except ErrorMessage as err_msg:
            raise ApiError(f'Result with id {result_id} does not exist!') from err_msg
        return dict(result)

    def get_results(self) -> List[Dict[str, Any]]:
        """ Gets all the results registered for the user the API is currently authenticated for.

        :return:
            The results with all of its properties.
            See :meth:`~.get_result` for a description of the result properties.
        """
        ret: List[Dict[str, Any]] = self._action(['results', 'list'])
        return ret

    def get_result_from_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the result with its properties for a single job, given the job id.

        :param job_id: The job identification number.

        :return:
            The result with its properties for the job with identification `job_id`.
            See :meth:`~.get_result` for a description of the result properties.

        :raises ApiError: If the job identified by `job_id` does not exist.
        """
        try:
            result = self._action(['jobs', 'result', 'list'], params={'id': job_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Job with id {job_id} does not exist!') from err_msg
        return dict(result)

    def get_raw_data_from_result(self, result_id: int) -> List[List[Any]]:
        """ Gets the raw data from the result.

        Gets the raw data from the result of the executed job, given the result_id.
        The raw data consists of a list of measurements for each shot (job.number_of_shots).
        The measurements contain a list of data for each measurement in the job (len(measurement_mask)).
        The data consist of integer state values (0 or 1) for each measured qubit, else None (see measurement_mask).
        Actual type is: List[List[List[Optional[int]]]] with least significant qubit first.

        Example for :code:`"measurement_mask": [[0, 1], [1, 1]]` the raw_data for 6 shots is structured as follows:

        .. code-block::

            "raw_data": [
                [[None, 1], [0, 1]],
                [[None, 1], [1, 0]],
                [[None, 0], [1, 1]],
                [[None, 0], [1, 1]],
                [[None, 0], [0, 0]],
                [[None, 1], [0, 1]]
            ]

        In this example in the first measurement block only qubit[1] is measured, in the second measurement block
        qubit[0] and qubit[1] are measured.

        :param result_id: The identification number of the result.

        :raises ApiError: If the raw data url in result is invalid or the request for the raw data using the url failed.

        :return:
            The raw data as a list of structures. A list with an empty list is returned when raw data has no elements.
        """
        result = self.get_result(result_id)
        raw_data_url = str(result.get('raw_data_url'))
        try:
            token = raw_data_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError(f'Invalid raw data url for result with id {result_id}!') from err_msg
        try:
            raw_data: List[List[Any]] = self._action(['results', 'raw-data', 'read'],
                                                     params={'id': result_id,
                                                             'token': token})
        except ErrorMessage as err_msg:
            raise ApiError(f'Raw data for result with id {result_id} does not exist!') from err_msg
        return raw_data

    def get_quantum_states_from_result(self, result_id: int) -> List[List[Any]]:
        """ Gets the quantum states for each measurement block from the result of the executed job, given the result_id.

        :param result_id: The identification number of the result.

        :raises ApiError: If the quantum states url in result is invalid or the request
            for the quantum states using the url failed.

        :return:
            A list of quantum states. The quantum states consists of a list of quantum state values and
            their corresponding probability.
            An empty list is returned when the backend does not return quantum states.
        """
        result = self.get_result(result_id)
        quantum_states_url = str(result.get('quantum_states_url'))
        try:
            token = quantum_states_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError(f'Invalid quantum states url for result with id {result_id}!') from err_msg
        try:
            quantum_states: List[List[Any]] = self._action(['results', 'quantum-states', 'read'],
                                                           params={'id': result_id, 'token': token})
        except ErrorMessage as err_msg:
            raise ApiError(f'Quantum states for result with id {result_id} does not exist!') from err_msg
        return quantum_states

    def get_measurement_register_from_result(self, result_id: int) -> List[List[Any]]:
        """ Gets the measurement register for each measurement block from the result of the executed code,
        given the result_id.

        :param result_id: The identification number of the result.

        :raises ApiError: If the measurement register url in result is invalid
            or the request for the measurement register using the url failed.

        :return:
            A list of measurement registers. The measurement register consists of a list of measurement register values.
            An empty list is returned when there is no data.
        """
        result = self.get_result(result_id)
        measurement_register_url = str(result.get('measurement_register_url'))
        try:
            token = measurement_register_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError(f'Invalid measurement register url for result with id {result_id}!') from err_msg
        try:
            measurement_register: List[List[Any]] = self._action(['results', 'measurement-register', 'read'],
                                                                 params={'id': result_id, 'token': token})
        except ErrorMessage as err_msg:
            raise ApiError(f'Measurement register for result with id {result_id} does not exist!') from err_msg
        return measurement_register

    def get_calibration_from_result(self, result_id: int) -> Optional[Any]:
        """ Gets the calibration information of the executed job, given the result_id.

        :param result_id: The identification number of the result.

        :raises ApiError: If the calibration url in result is invalid or the request
            for the calibration info using the url failed.

        :return:
            The calibration information from the backend. Calibration information may differ per backend type.
            None is returned when there is no calibration information (e.g. simulator backends).
        """
        calibration_info = None
        result = self.get_result(result_id)
        calibration_url = result.get('calibration')
        if calibration_url is not None:
            try:
                token = str(calibration_url).split('/')[-2]
            except IndexError as err_msg:
                raise ApiError(f'Invalid calibration url for result with id {result_id}!') from err_msg
            try:
                calibration_info = self._action(['calibration', 'read'], params={'id': token})
            except ErrorMessage as err_msg:
                raise ApiError(f'Calibration info for result with id {result_id} does not exist!') from err_msg
        return calibration_info

    #  assets  #

    def list_assets(self) -> None:
        """ Prints a list of the assets registered to the current user the API is authenticated for.

        For each asset the name, identification number and project identification number is printed."""
        assets = self.get_assets()
        for asset in assets:
            print(f'Asset name: {asset["name"]}, id: {asset["id"]}, (project_id: {asset["project_id"]})')

    def get_asset(self, asset_id: int) -> Dict[str, Any]:
        """ Gets the properties of the asset, given the asset_id.

        :param asset_id: The asset identification number.

        :raises ApiError: If the requested asset does not exist.

        :return:
            The properties describing the asset:

            ======================= =========== ==================================================================
            Key                     Type        Description
            ======================= =========== ==================================================================
            ``url``                 str         The url to get this asset.
            ``id``                  int         Unique id of this asset.
            ``name``                str         The name to get the asset.
            ``contentType``         str         The description of the content e.g. 'application/qasm' or
                                                'text/plain'.
            ``content``             str         The content itself.
                                                For example a cQASM program when linked to a job.
            ``project``             str         Url to get the project properties for which
                                                this asset was created.
            ``project_id``          int         The project id of the project for which this asset was created.
            ======================= =========== ==================================================================
        """
        try:
            asset = self._action(['assets', 'read'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Asset with id {asset_id} does not exist!') from err_msg
        return dict(asset)

    def get_assets(self) -> List[Dict[str, Any]]:
        """ Gets all the assets registered for the user the API is currently authenticated for.

        :return:
            The assets with all of its properties.
            See :meth:`~.get_asset` for a description of the asset properties.
        """
        ret: List[Dict[str, Any]] = self._action(['assets', 'list'])
        return ret

    def get_assets_from_project(self, project_id: int) -> List[Dict[str, Any]]:
        """ Gets the assets with its properties for a single project, given the project id.

        :param project_id: The project identification number.

        :return:
            List of assets with its properties for the project with identification `project_id`.
            An empty list is returned when the project has no assets.

        :raises ApiError: If the project identified by `project_id` does not exist.
        """
        try:
            assets: List[Dict[str, Any]] = self._action(['projects', 'assets', 'list'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Project with id {project_id} does not exist!') from err_msg
        return assets

    def get_asset_from_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the asset data from the job, given the job_id.

        :param job_id: The identification number of the job.

        :raises ApiError: If the requested asset with identification from the job 'input' field does not exist.

        :return:
            The assets with all of its properties.
            See :meth:`~.get_asset` for a description of the asset properties.
        """
        job = self.get_job(job_id)
        asset_url = str(job.get('input'))
        try:
            asset_id = int(asset_url.split('/')[-2])
        except (ValueError, IndexError) as err_msg:
            raise ApiError(f'Invalid input url for job with id {job_id}!') from err_msg
        try:
            asset = self._action(['assets', 'read'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError(f'Asset with id {asset_id} does not exist!') from err_msg
        return dict(asset)

    def _create_asset(self, name: str, project: Dict[str, Any], content: str) -> Dict[str, Any]:
        """ Create an asset.

        This method is used by execute_qasm_async to generate a new asset with a unique name to hold the
        content of the cQASM program for the project given. Assets are deleted when the project for which the asset
        was created is deleted.

        :param name: The name of the new asset.
        :param project: The project to which the asset is linked to.
        :param content: The cQASM code content.

        :return:
            The properties of the asset.
            See :meth:`~.get_asset` for a description of the asset properties.
        """
        payload = {
            'name': name,
            'contentType': 'application/qasm',
            'project': project['url'],
            'content': content,
        }
        return dict(self._action(['assets', 'create'], params=payload))

    #  other  #

    def wait_for_completed_job(self, quantum_inspire_job: QuantumInspireJob, collect_max_tries: Optional[int] = None,
                               sec_retry_delay: float = 0.5) -> Tuple[bool, str]:
        """ Wait for job completion.

        Delays the process and requests the job status. The waiting loop is broken when the job status is
        completed or cancelled, or when the maximum number of tries is set and has been reached.

        :param quantum_inspire_job: A job object.
        :param collect_max_tries: The maximum number of times the job status is checked. When set, the value should
            be > 0. When not set, the method waits until the job status is either completed or cancelled.
        :param sec_retry_delay: The time delay in between job status checks in seconds.

        :return:
            True if the job result could be collected else False in hte first part of the tuple.
            The latter part of the tuple contains an (error)message.
        """
        attempts = itertools.count() if collect_max_tries is None else range(collect_max_tries)
        for _ in attempts:
            time.sleep(sec_retry_delay)
            status = quantum_inspire_job.check_status()
            if status == 'COMPLETE':
                return True, 'Job completed.'
            if status == 'CANCELLED':
                return False, 'Failed getting result: job cancelled.'
        return False, 'Failed getting result: timeout reached.'

    @staticmethod
    def _generate_error_result(message: str) -> Dict[str, Any]:
        """Generate an error result object

        :param message: Reason for the failed job

        :return:
            Result object containing an empty histogram and an error message
        """
        result_obj = {
            'histogram': [],
            'raw_text': message
        }
        return result_obj

    def execute_qasm(self, qasm: str, backend_type: Optional[Union[Dict[str, Any], int, str]] = None,
                     number_of_shots: Optional[int] = None, collect_tries: Optional[int] = None,
                     default_number_of_shots: Optional[int] = None, identifier: Optional[str] = None,
                     full_state_projection: bool = False, user_data: str = '') -> Dict[str, Any]:
        """ With this method a cQASM program is executed, and the result is returned when the job is completed.

        The method :meth:`~.execute_qasm_async` is called which returns a QuantumInspireJob directly without waiting
        for the job to complete. After this call a waiting loop is started to wait for the job to finish and get
        the result.
        When no project name was given when the QuantumInspireAPI was created, the job is linked to a newly
        created temporary project in :meth:`~.execute_qasm_async`. When the job has finished running, this project is
        deleted.

        Depending on how busy the backend is, it takes some time to execute the job and
        returning the result. This method waits for the job to finish. The parameter collect_tries defines the
        maximum waiting time (collect_tries x 0.5 seconds). When the job takes longer to finish no results
        are returned. When set, the value of collect_tries must be > 0. When collect_tries is not set,
        the waiting time for completion is not limited.

        :param qasm: The cQASM code as string object.
        :param backend_type: The backend_type to execute the algorithm on.
        :param number_of_shots: Execution times of the algorithm before collecting the results.
        :param collect_tries: The number of times the status of the job is checked for completion before returning.
        :param default_number_of_shots: The default used number of shots for the project.
        :param identifier: The identifier to generate names for the project, asset and job when necessary.
        :param full_state_projection: Do not use full state projection with simulations when set to False (default).
        :param user_data: Data that the user wants to pass along with the job.

        :return:
            The results of the executed cQASM if successful else an error result if
            the results could not be collected within the given number of tries or the job failed.
            See :meth:`~.get_result` for a description of the result properties.
        """
        delete_project_afterwards = self.project_name is None
        quantum_inspire_job = None
        try:
            quantum_inspire_job = self.execute_qasm_async(qasm, backend_type=backend_type,
                                                          number_of_shots=number_of_shots,
                                                          default_number_of_shots=default_number_of_shots,
                                                          identifier=identifier,
                                                          full_state_projection=full_state_projection,
                                                          user_data=user_data)

            has_results, message = self.wait_for_completed_job(quantum_inspire_job, collect_tries)
            return dict(quantum_inspire_job.retrieve_results()) if has_results else \
                dict(self._generate_error_result(message))
        except (CoreAPIException, TypeError, ValueError, ApiError) as err_msg:
            message = f'Error raised while executing qasm: {err_msg}'
            return dict(self._generate_error_result(message))
        finally:
            if delete_project_afterwards and quantum_inspire_job is not None:
                project_identifier = quantum_inspire_job.get_project_identifier()
                self.delete_project(project_identifier)

    def execute_qasm_async(self, qasm: str, backend_type: Optional[Union[Dict[str, Any], int, str]] = None,
                           number_of_shots: Optional[int] = None, default_number_of_shots: Optional[int] = None,
                           identifier: Optional[str] = None, full_state_projection: bool = False,
                           project: Optional[Dict[str, Any]] = None, job_name: Optional[str] = None,
                           user_data: str = '') -> QuantumInspireJob:
        """With this method a cQASM program (job) is scheduled to be executed asynchronously.

        The method returns
        directly without waiting for the job to complete, as opposed to method :meth:`~.execute_qasm` which waits for
        the job to finish and returns the result.

        To execute a cQASM program a job is scheduled on a backend of type `backend type` as given by the
        parameter backend_type. Currently, there are 3 backend types available:

            1. QX single-node simulator (default for anonymous accounts)
            2. QX single-node simulator SurfSara (advanced account credentials needed)
            3. QX multi-node simulator SurfSara (advanced account credentials needed)

        When no `backend_type` is given, the default backend type currently being `QX single-node simulator`, is used.

        The job has to be linked with a project before it can be scheduled to execute.

        When a project name was supplied when the QuantumInspireAPI was created (see :meth:`~.__init__`), the job
        is linked to an existing project with this project name. When no project exists with this name a
        project with this name is created.

        When no project name was given when the QuantumInspireAPI was created, the job is linked to the project
        that is given as an argument `project`. When this `project` argument is empty, a project is created.
        First a project name is generated using the parameter `identifier` or when parameter `identifier` is
        empty an identifier is generated.

        When the project is created, it is created for the backend type and the default number of shots given.
        When the project already existed the values of the existing project are used for backend type and
        default number of shots.

        An asset with a unique id is created containing the cQASM program. This asset is linked to the project.

        When the project and the asset containing the program are known, a job is created with the name given by
        parameter `job_name`. When this parameter `job_name` is not filled, a job name is generated.

        The job that is created for running the program (contained in the asset) is linked to the project and will
        be executed `number_of_shots` times (as given by the parameter) before the results can be collected.

        The jobs' `user_data` is filled with the user data given as a parameter. This user data can be fetched and
        used later in the process.

        The default value of job parameter `full_state_projection` is set to False which
        means that the algorithm is treated as non-deterministic. As a result a deterministic algorithm may take
        longer to execute than strictly needed. When `full_state_projection` is set to True, a non-deterministic
        algorithm may give wrong results. Parameter `full_state_projection` is only used for simulations.

        Once the job is created, the method returns directly without waiting for the job to complete.
        The job is returned as a :class:`QuantumInspireJob`.
        This class encapsulates the job and contains methods the get
        the status of the job and retrieve the execution results when the job is completed.

        :param qasm: The qasm code as a string object.
        :param backend_type: The backend_type to execute the algorithm on.
        :param number_of_shots: Execution times of the algorithm before the results can be collected.
        :param default_number_of_shots: The default used number of shots for the project.
        :param identifier: The identifier used for generating names for the project, asset and job.
        :param full_state_projection: Do not use full state projection when set to False (default).
        :param project: The properties of an existing project, the asset and job are linked to. Only used
            when the project_name member of the api is empty.
        :param job_name: Name for the job that is to be executed, when None a job name is generated (see identifier)
        :param user_data: Data that the user wants to pass along with the job.

        :return:
            An encapsulated job object containing methods the get the status of the job and
            retrieve the execution results.
        """
        if not isinstance(backend_type, dict):
            if backend_type is None:
                backend_type = self.get_backend_type()
            elif isinstance(backend_type, int):
                backend_type = self.get_backend_type(int(backend_type))
            elif isinstance(backend_type, str):
                backend_type = self.get_backend_type(str(backend_type))

        if identifier is None:
            identifier = str(uuid.uuid1())

        if self.project_name is not None:
            project = next((project for project in self.get_projects()
                            if project['name'] == self.project_name), None)

        if project is None:
            if default_number_of_shots is None:
                default_number_of_shots = backend_type['default_number_of_shots']
            project_name = self.project_name if self.project_name else f'qi-sdk-project-{identifier}'
            project = self.create_project(project_name, default_number_of_shots, backend_type)

        if backend_type['url'] != project['backend_type']:
            logger.warning("The backend for which the project was created is different "
                           "from the backend type given: %s. The experiment is run on backend %s.",
                           backend_type['name'], backend_type['name'])

        qasm = qasm.lstrip()
        qasm = re.sub(r'[ \t]*\n[ \t]*', r'\n', qasm)
        asset_name = f'qi-sdk-asset-{identifier}'
        asset = self._create_asset(asset_name, project, qasm)

        if job_name is None:
            job_name = f'qi-sdk-job-{identifier}'
        if number_of_shots is None:
            number_of_shots = backend_type['default_number_of_shots']
        job = self._create_job(job_name, asset, number_of_shots, backend_type, user_data=user_data,
                               full_state_projection=full_state_projection)

        return QuantumInspireJob(self, job['id'])



# Quantum Inspire SDK
#
# Copyright 2022 QuTech Delft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module job
==========

.. autoclass:: QuantumInspireJob
   :members:
"""

from __future__ import annotations

from typing import Dict, Any, TYPE_CHECKING
from coreapi.exceptions import ErrorMessage

if TYPE_CHECKING:
    from api import QuantumInspireAPI


class QuantumInspireJob:

    def __init__(self, api: QuantumInspireAPI, job_identifier: int) -> None:
        """ Encapsulation of a job.

        The :py:class:`QuantumInspireJob` class encapsulates the base job of the API and has
        methods to check the status and retrieve the results from the API.

        :param api: An instance to the API.
        :param job_identifier: The job identification number.
        """
        QuantumInspireJob.__check_arguments(api, job_identifier)
        self.__job_identifier: int = job_identifier
        self.__api: QuantumInspireAPI = api

    @staticmethod
    def __check_arguments(api: QuantumInspireAPI, job_identifier: int) -> None:
        """ Checks whether the supplied arguments are of correct type.

        :param api: An instance to the API.
        :param job_identifier: The job identification number.

        :raises ValueError: When the api is not a :py:class:`QuantumInspireAPI` or when the
            job identifier is not found.
        """
        if type(api).__name__ != 'QuantumInspireAPI':
            raise ValueError('Invalid Quantum Inspire API!')
        try:
            _ = api.get_job(job_identifier)
        except ErrorMessage as error:
            raise ValueError('Invalid job identifier!') from error

    def check_status(self) -> str:
        """ Checks the execution status of the job.

        :return:
            The status of the job. Can be: 'NEW', 'RUNNING', 'COMPLETE', 'CANCELLED'
        """
        job = self.__api.get_job(self.__job_identifier)
        return str(job['status'])

    def retrieve_results(self) -> Dict[str, Any]:
        """ Gets the results of the job.

        :return:
            The execution results with a histogram item containing the result
            histogram of the job. When an error has occurred the raw_text item shall not be
            an empty string.
        """
        result: Dict[str, Any] = self.__api.get_result_from_job(self.__job_identifier)
        return result

    def get_job_identifier(self) -> int:
        """ Gets the set job identification number for the wrapped job.

        :return:
            The job identification number.
        """
        return self.__job_identifier

    def get_project_identifier(self) -> int:
        """ Gets the project identification number of the wrapped job.

        :return:
            The project identification number.
        """
        asset = self.__api.get_asset_from_job(self.__job_identifier)
        return int(asset['project_id'])
